import os
import inspect
from functools import cached_property
from abc import ABC, abstractmethod
import numpy as np
import shapefile
from pyproj import Proj, Geod
from matplotlib import colormaps
from NEDAS.utils.graphics import draw_line, draw_patch, arrowhead_xy, draw_reference_vector_legend

class GridBase(ABC):
    """
    Base class to handle 2D fields defined on regular grids or unstructured meshes.

    Args:
        proj (pyproj.Proj, custom func, None):
            Projection function mapping from longitude,latitude to x,y coordinates.
            If None, a default Mercator projection will be used.
        x (np.ndarray): X-coordinates for each grid point.
        y (np.ndarray): Y-coordinates for each grid point.
        bounds (list, optional):
            Grid boundary limits, [xmin, xmax, ymin, ymax], all float numbers.
            If not specified, will use min/max value of the coordinates.
        regular (bool, optional):
            Whether grid is regular or unstructured. Default is True (regular grid).
        cyclic_dim (str, optional):
            Cyclic dimension(s): ``'x'``, ``'y'``, ``'xy'``, or ``None`` if noncyclic.
        distance_type (str, optional): Type of distance functions: `cartesian` (default) or `spherical`.
        dst_grid (GridBase, optional): Destination grid object to convert to.

    Attributes:
        proj (pyproj.Proj or custom function): Projection function.
        proj_name (str): Name of the projection, empty if not available.
        bounds (list): Grid boundary limits [xmin, xmax, ymin, ymax].
        mask (np.ndarray):
            Mask (bool) for points that are not participating the analysis,same shape as :code:`x`, default is all False.
    """
    def __init__(self, proj, x, y, bounds=None, cyclic_dim=None, distance_type='cartesian', dst_grid=None):
        assert x.shape == y.shape, "x, y shape does not match"

        if proj is None:
            self.proj = Proj('+proj=stere') ##default projection
        else:
            self.proj = proj

        ##name of the projection
        if hasattr(proj, 'name'):
            self.proj_name = proj.name
        else:
            self.proj_name = ''

        ##proj info, ellps is used in Geod for distance calculation
        self.proj_ellps = 'WGS84'
        self.proj_lon0 = 0
        self.proj_lat0 = 0
        if hasattr(proj, 'definition'):
            for e in proj.definition.split():
                es = e.split('=')
                if es[0]=='ellps':
                    self.proj_ellps = es[1]
                if es[0]=='lat_0':
                    self.proj_lat0 = np.float32(es[1])
                if es[0]=='lon_0':
                    self.proj_lon0 = np.float32(es[1])

        ##coordinates and properties of the 2D grid
        self.x = x
        self.y = y
        self.cyclic_dim = cyclic_dim

        ##internally we use -180:180 convention for longitude
        if self.proj_name == 'longlat':
            self.x = np.mod(self.x + 180., 360.) - 180.

        ##boundary corners of the grid
        if bounds is not None:
            self.xmin, self.xmax, self.ymin, self.ymax = bounds
        else:
            self.xmin = np.min(self.x)
            self.xmax = np.max(self.x)
            self.ymin = np.min(self.y)
            self.ymax = np.max(self.y)

        self.distance_type = distance_type

        self.mask = np.full(self.x.shape, False)

        self._dst_grid = None
        if dst_grid is not None:
            self.set_destination_grid(dst_grid)

    def __eq__(self, other):
        if not isinstance(other, GridBase):
            return False
        if self.proj != other.proj:
            return False
        if self.x.shape != other.x.shape:
            return False
        if not np.allclose(self.x, other.x):
            return False
        if not np.allclose(self.y, other.y):
            return False
        return True

    @property
    def mfx(self):
        """
        Map scaling factors in x direction (mfx), since on the projection plane dx is not exactly
        the distance on Earth. The mfx is defined as ratio between dx and the actual distance.
        """
        if self.proj_name == 'longlat':
            ##long/lat grid doesn't have units in meters, so will not use map factors
            return np.ones(self.x.shape)
        else:
            ##map factor: ratio of (dx, dy) to their actual distances on the earth.
            geod = Geod(ellps=self.proj_ellps)
            lon, lat = self.proj(self.x, self.y, inverse=True)
            lon1x, lat1x = self.proj(self.x+self.dx, self.y, inverse=True)
            _,_,gcdx = geod.inv(lon, lat, lon1x, lat1x)
            return self.dx / gcdx

    @property
    def mfy(self):
        """
        Map scaling factors in y direction (mfy), since on the projection plane dy is not exactly
        the distance on Earth. The mfy is defined as ratio between dy and the actual distance.
        """
        if self.proj_name == 'longlat':
            ##long/lat grid doesn't have units in meters, so will not use map factors
            return np.ones(self.x.shape)
        else:
            ##map factor: ratio of (dx, dy) to their actual distances on the earth.
            geod = Geod(ellps=self.proj_ellps)
            lon, lat = self.proj(self.x, self.y, inverse=True)
            lon1y, lat1y = self.proj(self.x, self.y+self.dy, inverse=True)
            _,_,gcdy = geod.inv(lon, lat, lon1y, lat1y)
            return self.dy / gcdy

    @property
    def dst_grid(self):
        """
        Destination grid for convert, interp, rotate_vector methods
        once specified a dst_grid, the setter will compute corresponding rotation_matrix and interp_weights
        """
        return self._dst_grid

    @dst_grid.setter
    def dst_grid(self, grid):
        assert isinstance(grid, GridBase), "dst_grid should be a Grid instance"
        if grid == self.dst_grid:  ##the same grid is set before
            return

        self._dst_grid = grid

        ##rotation of vector field from self.proj to dst_grid.proj
        self._set_rotation_matrix()

        ##prepare indices and weights for interpolation
        ##when dst_grid is set, these info are prepared and stored to avoid recalculating
        ##too many times, when applying the same interp to a lot of flds
        x, y = self._proj_from(grid.x, grid.y)
        inside, indices, vertices, in_coords, nearest = self.find_index(x, y)
        self.interp_inside = inside
        self.interp_indices = indices
        self.interp_vertices = vertices
        self.interp_nearest = nearest
        self.interp_weights = self._interp_weights(inside, vertices, in_coords)

        ##prepare indices for coarse-graining
        x, y = self._proj_to(self.x, self.y)
        inside, _, _, _, nearest = self.dst_grid.find_index(x, y)
        self.coarsen_inside = inside
        self.coarsen_nearest = nearest
        if not self.regular: ## for irregular mesh, find indices for elements too
            x, y = self._proj_to(self.x_elem, self.y_elem)
            inside, _, _, _, nearest = self.dst_grid.find_index(x, y)
            self.coarsen_inside_elem = inside
            self.coarsen_nearest_elem = nearest

    def set_destination_grid(self, grid):
        """
        Set method for self.dst_grid the destination Grid object to convert to.
        """
        self.dst_grid = grid

    def _wrap_cyclic_xy(self, x_, y_):
        """
        When interpolating for point x_,y_, if the coordinates falls outside of the domain,
        we wrap around and make then inside again, if the boundary condition is cyclic (self.cyclic_dim)

        Args:
            x_, y_ (np.array): x, y coordinates of a grid

        Returns:
            Same as input but x, y values are wrapped to be within the boundaries again.
        """
        if self.cyclic_dim is not None:
            for d in self.cyclic_dim:
                if d=='x':
                    x_ = np.mod(x_ - self.xmin, self.Lx) + self.xmin
                elif d=='y':
                    y_ = np.mod(y_ - self.ymin, self.Ly) + self.ymin
        return x_, y_

    @abstractmethod
    def find_index(self, x_, y_):
        """
        Find indices of `self.x`, `self.y` corresponding to the given `x_`, `y_`.

        Args:
            x_ (float or np.ndarray): x-coordinates of target point(s).
            y_ (float or np.ndarray): y-coordinates of target point(s).

        Outputs:
            inside (np.ndarray of bool): Boolean array of shape `(x_.size,)` indicating whether
                each `x_`, `y_` point lies inside the grid.

            indices (np.ndarray of int or None): Indices of grid elements containing the input points.
                - For regular grids, this is `None` since vertices suffice to locate the grid box.
                - For unstructured meshes, these are indices into `tri.triangles`, from `tri_finder`.

            vertices (np.ndarray of int): Array of shape `(inside_size, n)`, where
                `n = 4` for regular grid boxes or `n = 3` for mesh triangles.
                These are indices into `self.x`, `self.y` (flattened) for the vertices
                of the grid element that each point falls in.

            in_coords (np.ndarray of float): Array of shape `(inside_size, n)` giving internal coordinates
                of each point within the containing element. Used to compute interpolation weights.

            nearest (np.ndarray of int): Array of shape `(inside_size,)` with indices of the grid nodes
                closest to each point.

        Notes:
            - This function assumes `self.x`, `self.y` define either a regular or triangular grid.
            - Internal coordinates are used for interpolation and vary in dimension based on the grid type.
        """
        pass

    def _proj_to(self, x, y):
        """
        Transform coordinates from self.proj to dst_grid.proj
        """
        if self.dst_grid.proj != self.proj:
            lon, lat = self.proj(x, y, inverse=True)
            x, y = self.dst_grid.proj(lon, lat)
        x, y = self.dst_grid._wrap_cyclic_xy(x, y)
        return x, y

    def _proj_from(self, x, y):
        """
        transform coordinates from dst_grid.proj to self.proj
        """
        if self.dst_grid.proj != self.proj:
            lon, lat = self.dst_grid.proj(x, y, inverse=True)
            x, y = self.proj(lon, lat)
        x, y = self._wrap_cyclic_xy(x, y)
        return x, y

    def _set_rotation_matrix(self):
        """
        setting the rotation matrix for converting vector fields from self to dst_grid
        """
        self.rotate_matrix = np.zeros((4,)+self.x.shape)
        if self.proj != self.dst_grid.proj:
            ##self.x,y corresponding coordinates in dst_proj, call them x,y
            x, y = self._proj_to(self.x, self.y)

            ##find small increments in x,y due to small changes in self.x,y in dst_proj
            eps = 0.1 * self.dx    ##grid spacing is specified in Grid object
            xu, yu = self._proj_to(self.x + eps, self.y      )  ##move a bit in x dirn
            xv, yv = self._proj_to(self.x      , self.y + eps)  ##move a bit in y dirn

            np.seterr(invalid='ignore')  ##will get nan at poles due to singularity, fill_pole_void takes care later
            dxu = xu-x
            dyu = yu-y
            dxv = xv-x
            dyv = yv-y
            hu = np.hypot(dxu, dyu)
            hv = np.hypot(dxv, dyv)
            self.rotate_matrix[0, :] = dxu/hu
            self.rotate_matrix[1, :] = dxv/hv
            self.rotate_matrix[2, :] = dyu/hu
            self.rotate_matrix[3, :] = dyv/hv
        else:
            ##if no change in proj, we can skip the calculation
            self.rotate_matrix[0, :] = 1.
            self.rotate_matrix[1, :] = 0.
            self.rotate_matrix[2, :] = 0.
            self.rotate_matrix[3, :] = 1.

    def rotate_vectors(self, vec_fld):
        """
        Apply the rotate_matrix to a vector field

        Args:
            vec_fld (np.array): The input vector field, shape (2, self.x.shape)

        Returns:
            The vector field rotated to the dst_grid.
        """
        u = vec_fld[0, :].copy()
        v = vec_fld[1, :].copy()
        rw = self.rotate_matrix
        u_rot = rw[0, :]*u + rw[1, :]*v
        v_rot = rw[2, :]*u + rw[3, :]*v
        return np.array([u_rot, v_rot])

    @abstractmethod
    def interp(self, fld, x=None, y=None, method='linear'):
        """
        Interpolation of 2D field data (fld) from one grid (self or given x,y) to another (dst_grid).
        This can be used for grid refining (low->high resolution) or grid thinning (high->low resolution).
        This also converts between different grid geometries.

        Args:
            fld (np.array): Input field defined on self, should have same shape as self.x
            x,y (float or np.array): Optional;
                If x,y are specified, the function computes the weights and apply them to fld
                If x,y are None, the self.dst_grid.x,y are used.
                Since their interp_weights are precalculated by dst_grid.setter it will be efficient
                to run interp for many different input flds quickly.
            method (str): Interpolation method, can be 'nearest' or 'linear'

        Returns:
            The interpolated field defined on the destination grid
        """
        pass

    @abstractmethod
    def coarsen(self, fld):
        """
        Coarse-graining is sometimes needed when the dst_grid is at lower resolution than self.
        Since many points of self.x,y falls in one dst_grid box/element, it is better to
        average them to represent the field on the low-res grid, instead of interpolating
        only from the nearest points that will cause representation errors.

        Args:
            fld (np.array): Input field to perform coarse-graining on, it is defined on self.

        Returns:
            The coarse-grained field defined on self.dst_grid.
        """
        pass

    def convert(self, fld, is_vector=False, method='linear', coarse_grain=False):
        """
        Main method to convert from self.proj, x, y to dst_grid coordinate systems:

        Notes:
            1. if projection changes and is_vector, rotate vectors from self.proj to dst_grid.proj
            2.1 interpolate fld components from self.x,y to dst_grid.x,y
            2.2 if dst_grid is low-res, coarse_grain=True will perform coarse-graining

        Args:
            fld (np.array): Input field to perform convertion on.
            is_vector (bool, optional):
                If False (default) the input fld is a scalar field,
                otherwise the input fld is a vector field.
            method (str, optional):
                Interpolation method, 'linear' (default) or 'nearest' 
            coarse_grain (bool, optional):
                If True, the coarse-graining will be applied using self.coarsen(). The default is False.

        Returns:
            The converted field defined on the destination grid self.dst_grid.
        """
        if self.dst_grid != self:
            if is_vector:
                assert fld.shape[0] == 2, "vector field should have first dim==2, for u,v component"
                ##vector field needs to rotate to dst_grid.proj before interp
                fld = self.rotate_vectors(fld)

                fld_out = np.full((2,)+self.dst_grid.x.shape, np.nan)
                for i in range(2):
                    ##interp each component: u, v
                    fld_out[i, :] = self.interp(fld[i, :], method=method)
                    if coarse_grain:
                        ##coarse-graining if more points fall in one grid
                        fld_coarse = self.coarsen(fld[i, :])
                        ind = ~np.isnan(fld_coarse)
                        fld_out[i, ind] = fld_coarse[ind]
            else:
                ##scalar field, just interpolate
                fld_out = np.full(self.dst_grid.x.shape, np.nan)
                fld_out = self.interp(fld, method=method)
                if coarse_grain:
                    ##coarse-graining if more points fall in one grid
                    fld_coarse = self.coarsen(fld)
                    ind = ~np.isnan(fld_coarse)
                    fld_out[ind] = fld_coarse[ind]
        else:
            fld_out = fld
        return fld_out

    def distance(self, ref_x, x, ref_y, y, p=2, type='cartesian'):
        """
        Compute distance for points (x,y) to the reference point
        
        Args:
            ref_x, ref_y (float):
                reference point x,y coordinates
            x, y (np.array):
                points whose distance to the reference points will be computed
            p (int, optional):
                Minkowski p-norm order, default is 2
            type (str, optional):
                distance type, 'cartesian' (default) or 'spherical'
        
        Returns:
            Distances between x,y and the reference point ref_x, ref_y.
        """
        if type == 'cartesian':
            ##normal cartesian distances in x and y
            dist_x = np.abs(x - ref_x)
            if self.cyclic_dim is not None and 'x' in self.cyclic_dim:
                dist_x = np.minimum(dist_x, self.Lx - dist_x)
            dist_y = np.abs(y - ref_y)
            if self.cyclic_dim is not None and 'y' in self.cyclic_dim:
                dist_y = np.minimum(dist_y, self.Ly - dist_y)
            if p == 1:
                dist = dist_x + dist_y  ##Manhattan distance, order 1
            elif p == 2:
                dist = np.hypot(dist_x, dist_y)   ##Euclidean distance, order 2
            else:
                raise NotImplementedError(f"grid.distance: p-norm order {p} is not implemented for 2D grid")
            return dist

        ##compute spherical distance on Earth instead
        elif type == 'spherical':
            reflon, reflat = self.proj(ref_x, ref_y, inverse=True)
            lon, lat = self.proj(x, y, inverse=True)
            RE = 6371000.0
            invrad = np.pi / 180.
            rlon1 = np.atleast_1d(reflon) * invrad
            rlat1 = np.atleast_1d(reflat) * invrad
            rlon2 = np.atleast_1d(lon) * invrad
            rlat2 = np.atleast_1d(lat) * invrad
            ##from m_spherdist.F90 in enkf-topaz:
            cos_d = np.sin(rlat1) * np.sin(rlat2) + np.cos(rlat1) * np.cos(rlat2) * np.cos(rlon1 - rlon2)
            dist = RE * np.acos(np.clip(cos_d, -1, 1))
            ##Haversine formula to avoid precision loss
            # a = np.sin((rlat2 - rlat1) / 2)**2 + np.cos(rlat1) * np.cos(rlat2) * np.sin((rlon1 - rlon2) / 2)**2
            # dist = 2 * RE * np.asin(np.sqrt(a))
            return dist

        else:
            raise ValueError(f"unknown distance type '{type}'")

    def _collect_shape_data(self, shapes):
        """
        This collects the x,y coordinates from shapes read from .shp files for later plotting
        filter the points not inside the grid domain
        """
        data = {'xy':[], 'parts':[]}
        for shape in shapes:
            if len(shape.points) > 0:
                xy = []
                inside = []
                lon, lat = [np.array(x) for x in zip(*shape.points)]
                x, y = self.proj(lon, lat)
                inside = np.logical_and(np.logical_and(x >= self.xmin, x <= self.xmax),
                                        np.logical_and(y >= self.ymin, y <= self.ymax))

                ##when showing global maps, the lines leave the domain and re-enter
                ##from the other side, the cross-over lines are visible on the plot
                ##temporary solution: make a pause when lines wrap around cut meridian
                ##lines work fine now but filled patches do not
                if self.proj_name in ['longlat', 'tripolar', 'bipolar']:
                    x[~inside] = np.nan
                    y[~inside] = np.nan

                xy = [(x[i], y[i]) for i in range(x.size)]

                ##if any point in the polygon lies inside the grid, need to plot it.
                if any(inside):
                    data['xy'].append(xy)
                    data['parts'].append(shape.parts)

        return data

    @cached_property
    def land_data(self):
        """
        prepare data to show the land area, the shp file ne_50m_coastlines is
        downloaded from https://www.naturalearthdata.com
        """
        path = os.path.split(inspect.getfile(self.__class__))[0]
        sf = shapefile.Reader(os.path.join(path, 'ne_50m_coastline.shp'))
        shapes = sf.shapes()

        ##Some cosmetic tweaks of the shapefile for some Canadian coastlines
        shapes[1200].points = shapes[1200].points + shapes[1199].points[1:]
        shapes[1199].points = []
        shapes[1230].points = shapes[1230].points + shapes[1229].points[1:] + shapes[1228].points[1:] + shapes[1227].points[1:]
        shapes[1229].points = []
        shapes[1228].points = []
        shapes[1227].points = []
        shapes[1233].points = shapes[1233].points + shapes[1234].points
        shapes[1234].points = []
        shapes[1234].points = []

        return self._collect_shape_data(shapes)

    @cached_property
    def river_data(self):
        """
        prepare data to show river features
        """
        path = os.path.split(inspect.getfile(self.__class__))[0]
        sf = shapefile.Reader(os.path.join(path, 'ne_50m_rivers.shp'))
        shapes = sf.shapes()
        return self._collect_shape_data(shapes)

    @cached_property
    def lake_data(self):
        """
        prepare data to show lake features
        """
        path = os.path.split(inspect.getfile(self.__class__))[0]
        sf = shapefile.Reader(os.path.join(path, 'ne_50m_lakes.shp'))
        shapes = sf.shapes()
        return self._collect_shape_data(shapes)

    def llgrid_xy(self, dlon:float, dlat:float):
        """
        Prepare a lon/lat grid to plot as reference lines

        Args:
        - dlon, dlat: spacing of lon/lat grid in degrees
        """
        self.dlon = dlon
        self.dlat = dlat

        llgrid_xy = []
        for lon_r in np.arange(-180, 180, dlon):
            xy = []
            inside = []
            lat = np.arange(-89.9, 90, 0.1)
            lon = np.ones(lat.size) * lon_r
            x, y = self.proj(lon, lat)
            inside = np.logical_and(np.logical_and(x >= self.xmin, x <= self.xmax),
                                    np.logical_and(y >= self.ymin, y <= self.ymax))
            x[~inside] = np.nan
            y[~inside] = np.nan
            xy = [(x[i], y[i]) for i in range(x.size)]
            if any(inside):
                llgrid_xy.append(xy)

        for lat_r in np.arange(-90, 90+dlat, dlat):
            xy = []
            inside = []
            lon = np.arange(-180., 180., 0.1)
            lat = np.ones(lon.size) * lat_r
            x, y = self.proj(lon, lat)
            inside = np.logical_and(np.logical_and(x >= self.xmin, x <= self.xmax),
                                    np.logical_and(y >= self.ymin, y <= self.ymax))
            x[~inside] = np.nan
            y[~inside] = np.nan
            xy = [(x[i], y[i]) for i in range(x.size)]
            if any(inside):
                llgrid_xy.append(xy)

        return llgrid_xy

    @abstractmethod
    def plot_field(self, ax, fld, vmin=None, vmax=None, cmap='viridis', **kwargs):
        """
        Plot a scalar field using pcolor/tripcolor

        Args:
            ax (matplotlib.pyplot.Axes): Axes handle for plotting
            fld (np.array): The scalar field for plotting
            vmin, vmax (float, optional):
                The minimum and maximum value range for the colormap, if not specified (None)
                the np.min, np.max of the input fld will be used.
            cmap (matplotlib colormap, or str, optional):
                Colormap used in the plot, default is 'viridis'
        """
        pass

    def plot_vectors(self, ax, vec_fld, V=None, L=None, spacing=0.5, num_steps=10,
                     linecolor='k', linewidth=1,
                     showref=False, ref_xy=(0.9, 0.9), refcolor='w', ref_units='',
                     showhead=True, headwidth=0.1, headlength=0.3):
        """
        Plot vector fields (improved version of matplotlib quiver)

        Args:
            ax (matplotlib.pyplot.Axes): Axes handle for plotting
            vec_fld (np.array): The vector field for plotting
            V (float, optional):
                Velocity scale, typical velocity value in vec_fld units. If not specified (None)
                a typical value 0.33*max(abs(vec_fld[0,:])) will be used.
            L (float, optional):
                Length scale, how long in x,y units do vectors with velocity V show in the plot
                If not specified (None), a typical value 0.05*self.Lx will be used.
            spacing (float, optional):
                Distance between vectors in both directions is given by spacing*L. Default is 0.5.
                This controls the density of vectors in the plot.
                You can provide a tuple (float, float) for spacings in (x, y) if you want them
                to be set differently.
            num_steps (int, optional):
                Default is 10. If num_steps=1, straight vectors (as in quiver) will be displayed.
                num_steps>1 lets you display curved trajectories, at each sub-step the velocity is
                re-interpolated at the new position along the trajectories. As num_steps get larger
                the trajectories are more detailed.
            linecolor (str or matplotlib color, optional):
                Line color for the vector lines, default is 'k'
            linewidth (float, optional):
                Line width for the vector lines, default is 1.
            showref (bool, optional):
                If True, show a legend box with a reference vector (size L) inside. Default is False.
            ref_xy (tuple, optional):
                The x,y relative coordinates (0-1) for the reference vector box, default is upper right corner.
            ref_color (str or matplotlib color, optional):
                Background color for the reference vector box, default is 'w' (white).
            ref_units (str, optional):
                Units to be included in the reference vector box, default is ''.
            showhead (bool, optional):
                If True (default), show the arrow head of the vectors
            headwidth (float, optional):
                Width of arrow heads relative to L, default is 0.1.
            headlength (float, optional):
                Length of arrow heads relative to L, default is 0.3.
        """
        assert vec_fld.shape == (2,)+self.x.shape, "vector field shape mismatch with x,y"
        x = self.x
        y = self.y
        u = vec_fld[0,:]
        v = vec_fld[1,:]

        ##set typicall L, V if not defined
        if V is None:
            V = 0.33 * np.nanmax(np.abs(u))
        if L is None:
            L = 0.05 * (np.max(x) - np.min(x))

        ##start trajectories on a regular grid with spacing d
        if isinstance(spacing, tuple):
            d = (spacing[0]*L, spacing[1]*L)
        else:
            d = (spacing*L, spacing*L)

        dt = L / V / num_steps

        xo, yo = np.mgrid[x.min()+0.5*d[0]:x.max():d[0], y.min()+0.5*d[1]:y.max():d[1]]
        npoints = xo.flatten().shape[0]
        xtraj = np.full((npoints, num_steps+1,), np.nan)
        ytraj = np.full((npoints, num_steps+1,), np.nan)
        leng = np.zeros(npoints)
        xtraj[:, 0] = xo.flatten()
        ytraj[:, 0] = yo.flatten()

        for t in range(num_steps):
            ###find velocity ut,vt at traj position for step t
            ut = self.interp(u, xtraj[:,t], ytraj[:,t])
            vt = self.interp(v, xtraj[:,t], ytraj[:,t])

            ###velocity should be in physical units, to plot the right length on projection
            ###we use the map factors to scale distance units
            ut = ut * self.interp(self.mfx, xtraj[:,t], ytraj[:,t])
            vt = vt * self.interp(self.mfy, xtraj[:,t], ytraj[:,t])

            ###update traj position
            xtraj[:, t+1] = xtraj[:, t] + ut * dt
            ytraj[:, t+1] = ytraj[:, t] + vt * dt

            ##update length
            leng = leng + np.hypot(ut, vt) * dt

        ##plot the vector lines
        hl = headlength * L
        hw = headwidth * L
        for i in range(xtraj.shape[0]):
            ##plot trajectory at one output location
            ax.plot(xtraj[i, :], ytraj[i, :], color=linecolor, linewidth=linewidth, zorder=4)
            ##add vector head if traj is long and straight enough
            dist = np.hypot(xtraj[i,0]-xtraj[i,-1], ytraj[i,0]-ytraj[i,-1])
            if showhead and hl < leng[i] < 1.6*dist:
                ax.fill(*arrowhead_xy(xtraj[i,-1], xtraj[i,-2], ytraj[i,-1],ytraj[i,-2], hw, hl), color=linecolor, zorder=5)

        ##add reference vector
        if showref:
            xr, yr = ref_xy[0]*self.Lx+self.xmin, ref_xy[1]*self.Ly+self.ymin
            draw_reference_vector_legend(ax, xr, yr, V, L, hw, hl, refcolor, linecolor, ref_units)

        self.set_xylim(ax)

    def plot_scatter(self, ax, fld, vmin=None, vmax=None, nlevels=20, cmap='viridis', markersize=10, x=None, y=None, is_vector=False, **kwargs):
        """
        Same as plot_field/vectors, but showing individual scattered points instead
        This is more suitable for plotting observations in space
        """
        if x is None:
            x = self.x
        if y is None:
            y = self.y

        if vmin is None:
            vmin = np.nanmin(fld)
        if vmax is None:
            vmax = np.nanmax(fld)
        dv = (vmax - vmin) / nlevels

        if is_vector:
            assert fld.shape[0] == 2
            assert fld.shape[1:] == x.shape
            V = vmax
            L = 0.05 * self.Lx
            hl, hw = 0.3 * L, 0.15 * L
            ref_x, ref_y = 0.9 * self.Lx + self.xmin, 0.9 * self.Ly + self.ymin
            refcolor = kwargs.get('refcolor', 'w')
            ref_units = kwargs.get('units', '')
            d = fld * L / V
            xtraj, ytraj = np.array([x, x + d[0,...]]), np.array([y, y + d[1,...]])
            linecolor = kwargs.get('linecolor', 'k')
            linewidth = kwargs.get('linewidth', 1)
            for i in np.ndindex(x.shape):
                ax.plot(xtraj[:,i], ytraj[:,i], color=linecolor, linewidth=linewidth, zorder=5)
                dist = np.hypot(xtraj[0,i]-xtraj[1,i], ytraj[0,i]-ytraj[1,i])
                if hl < 1.6*dist:
                    ax.fill(*arrowhead_xy(xtraj[1,i], xtraj[0,i], ytraj[1,i], ytraj[0,i], hw, hl), color=linecolor, zorder=5)
            draw_reference_vector_legend(ax, ref_x, ref_y, V, L, hw, hl, refcolor, linecolor, ref_units)

        else:
            assert fld.shape == x.shape
            msk = ~np.isnan(fld)
            v = np.array(fld[msk])
            vbound = np.maximum(np.minimum(v, vmax), vmin)

            if isinstance(cmap, str):
                cmap = colormaps[cmap]
            cmap = np.array([cmap(x)[0:3] for x in np.linspace(0, 1, nlevels+1)])

            cind = ((vbound - vmin) / dv).astype(int)
            ax.scatter(x[msk], y[msk], markersize, color=cmap[cind], **kwargs)

        self.set_xylim(ax)

    def plot_land(self, ax, color=None, linecolor='k', linewidth=1,
                  showriver=False, rivercolor='c',
                  showgrid=True, dlon=20, dlat=5):
        """
        Shows the map (coastline, rivers, lakes) and lon/lat grid for reference

        Args:
            ax (matplotlib.pyplot.Axes): Axes handle for plotting
            color (matplotlib color, optional):
                Face color of the landmass polygon, default is None (transparent).
            linecolor (matplotlib color, optional):
                Line color of the coastline, default is 'k' (black).
            linewidth (float, optional):
                Line width of the coastline, default is 1.
            showriver (bool, optional):
                If True, show the rivers and lakes over the landmass. Default is False.
            rivercolor (matplotlib color, optional):
                Color of the rivers and lakes, default is 'c' (cyan).
            showgrid (bool, optional):
                If True (default), show the reference lat/lon grid.
            dlon (float, optional):
                The interval of longitude lines in the reference grid. Default is 20 degrees.
            dlat (float, optional):
                The interval in latitude lines in the reference grid. Default is 5 degress.
        """
        ###plot the coastline to indicate land area
        if color is not None:
            draw_patch(ax, self.land_data, color=color, zorder=3)
        if linecolor is not None:
            draw_line(ax, self.land_data, linecolor=linecolor, linewidth=linewidth, linestyle='-', zorder=8)
        if showriver:
            draw_line(ax, self.river_data, linecolor=rivercolor, linewidth=0.5, linestyle='-', zorder=1)
            draw_patch(ax, self.lake_data, color=rivercolor, zorder=1)

        ###add reference lonlat grid on map
        if showgrid:
            for xy in self.llgrid_xy(dlon, dlat):
                ax.plot(*zip(*xy), color='k', linewidth=0.5, linestyle=':', zorder=4)

        self.set_xylim(ax)

    def set_xylim(self, ax):
        ##set the correct extent of plot
        ax.set_xlim(self.xmin, self.xmax)
        ax.set_ylim(self.ymin, self.ymax)



import os
from ..time_format import forday, dayfor

def bool_str(value):
    if value:
        vstr = 'T'
    else:
        vstr = 'F'
    return vstr

def namelist(m, time, forecast_period, run_dir='.'):
    """
    Generate namelists for TP4 runs,
    Inputs:
    - m: Model object with configs
    - time: datetime obj for the starting time of the run
    - forecast_period: forecast period in hours for the run
    - run_dir: path to the run directory
    """
    nmlstr =  f"ECWMF forcing; flx-s14w; LWcorr; precip+2mm; SSSrlx; FCT2 tsadvc.; 0-tracer.\n"
    nmlstr += f"Sigma0; GDEM3 init; KPP mixed layer; SeaWiFS mon KPAR; nested in ATLd0.08 2.6;\n"
    nmlstr += f"S-Z(15-11): dp00/f/x/i=3m/1.125/12m/1m; ds=1m/1.125/4m; src_2.2.12;\n"
    nmlstr += f"12345678901234567890123456789012345678901234567890123456789012345678901234567890\n"
    nmlstr += f"{m.iversn:4d}      'iversn' = hycom version number x10\n"
    nmlstr += f" {m.iexpt:03d}      'iexpt ' = experiment number x10\n"
    nmlstr += f"{m.idm:4d}      'idm   ' = longitudinal array size\n"
    nmlstr += f"{m.jdm:4d}      'jdm   ' = latitudinal  array size\n"
    nmlstr += f"{m.itest:4d}      'itest ' = grid point where detailed diagnostics are desired\n"
    nmlstr += f"{m.jtest:4d}      'jtest ' = grid point where detailed diagnostics are desired\n"
    nmlstr += f"{m.kdm:4d}      'kdm   ' = number of layers\n"
    nmlstr += f"{m.nhybrd:4d}      'nhybrd' = number of hybrid levels (0=all isopycnal)\n"
    nmlstr += f"{m.nsigma:4d}      'nsigma' = number of sigma  levels (nhybrd-nsigma z-levels)\n"
    nmlstr += f"{str(m.dp00):^8s}  'dp00  ' = deep    z-level spacing minimum thickness (m)\n"
    nmlstr += f"{str(m.dp00x):^8s}  'dp00x ' = deep    z-level spacing maximum thickness (m)\n"
    nmlstr += f"{str(m.dp00f):^8s}  'dp00f ' = deep    z-level spacing stretching factor (1.0=const.space)\n"
    nmlstr += f"{str(m.ds00):^8s}  'ds00  ' = shallow z-level spacing minimum thickness (m)\n"
    nmlstr += f"{str(m.ds00x):^8s}  'ds00x ' = shallow z-level spacing maximum thickness (m)\n"
    nmlstr += f"{str(m.ds00f):^8s}  'ds00f ' = shallow z-level spacing stretching factor (1.0=const.space)\n"
    nmlstr += f"{str(m.dp00i):^8s}  'dp00i ' = deep iso-pycnal spacing minimum thickness (m)\n"
    nmlstr += f"{str(m.isotop):^8s}  'isotop' = shallowest depth for isopycnal layers     (m, <0 from file)\n"
    nmlstr += f"{str(m.saln0):^8s}  'saln0 ' = initial salinity value (psu), only used for iniflg<2\n"
    nmlstr += f"{str(m.locsig):^8s}  'locsig' = locally-referenced pot. density for stability (0=F,1=T)\n"
    nmlstr += f"{str(m.kapref):^8s}  'kapref' = thermobaric ref. state (-1=input,0=none,1,2,3=constant)\n"
    nmlstr += f"{str(m.thflag):^8s}  'thflag' = reference pressure flag (0=Sigma-0, 2=Sigma-2)\n"
    nmlstr += f"{str(m.thbase):^8s}  'thbase' = reference density (sigma units)\n"
    nmlstr += f"{str(m.vsigma):^8s}  'vsigma' = spacially varying isopycnal target densities (0=F,1=T)\n"
    for k in range(m.kdm):
        nmlstr += f"{str(m.sigma[k]):^8s}  'sigma ' = layer  {k+1}  density (sigma units)\n"
    nmlstr += f"{str(m.iniflg):^8s}  'iniflg' = initial state flag (0=levl, 1=zonl, 2=clim)\n"
    nmlstr += f"{str(m.jerlv0):^8s}  'jerlv0' = initial jerlov water type (1 to 5; 0 to use KPAR)\n"
    nmlstr += f"{str(m.yrflag):^8s}  'yrflag' = days in year flag   (0=360,  1=366,  2=366J1, 3=actual)\n"
    nmlstr += f"{str(m.sshflg):^8s}  'sshflg' = diagnostic SSH flag (0=SSH,1=SSH&stericSSH)\n"
    nmlstr += f"{str(m.dsurfq):^8s}  'dsurfq' = number of days between model diagnostics at the surface\n"
    nmlstr += f"{str(m.diagfq):^8s}  'diagfq' = number of days between model diagnostics\n"
    nmlstr += f"{str(m.proffq):^8s}  'proffq' = number of days between model diagnostics at selected locs\n"
    nmlstr += f"{str(m.tilefq):^8s}  'tilefq' = number of days between model diagnostics on selected tiles\n"
    nmlstr += f"{str(m.meanfq):^8s}  'meanfq' = number of days between model diagnostics (time averaged)\n"
    nmlstr += f"{str(m.rstrfq):^8s}  'rstrfq' = number of days between model restart output\n"
    nmlstr += f"{str(m.bnstfq):^8s}  'bnstfq' = number of days between baro nesting archive input\n"
    nmlstr += f"{str(m.nestfq):^8s}  'nestfq' = number of days between 3-d  nesting archive input\n"
    nmlstr += f"{str(m.cplifq):^8s}  'cplifq' = number of days (or time steps) between sea ice coupling\n"
    nmlstr += f"{str(m.baclin):^8s}  'baclin' = baroclinic time step (seconds), int. divisor of 86400\n"
    nmlstr += f"{str(m.batrop):^8s}  'batrop' = barotropic time step (seconds), int. div. of baclin/2\n"
    nmlstr += f"{str(m.incflg):^8s}  'incflg' = incremental update flag (0=no, 1=yes, 2=full-velocity)\n"
    nmlstr += f"{str(m.incstp):^8s}  'incstp' = no. timesteps for full update (1=direct insertion)\n"
    nmlstr += f"{str(m.incupf):^8s}  'incupf' = number of days of incremental updating input\n"
    nmlstr += f"{str(m.wbaro ):^8s}  'wbaro ' = barotropic time smoothing weight\n"
    nmlstr += f"{str(m.btrlfr):^8s}  'btrlfr' = leapfrog barotropic time step (0=F,1=T)\n"
    nmlstr += f"{str(m.btrmas):^8s}  'btrmas' = barotropic is mass conserving (0=F,1=T)\n"
    nmlstr += f"{str(m.hybrlx):^8s}  'hybrlx' = HYBGEN: inverse relaxation coefficient (time steps)\n"
    nmlstr += f"{str(m.hybiso):^8s}  'hybiso' = HYBGEN: Use PCM if layer is within hybiso of target density\n"
    nmlstr += f"{str(m.hybmap):^8s}  'hybmap' = hybrid   remapper  flag (0=PCM, 1=PLM,  2=PPM)\n"
    nmlstr += f"{str(m.hybflg):^8s}  'hybflg' = hybrid   generator flag (0=T&S, 1=th&S, 2=th&T)\n"
    nmlstr += f"{str(m.advflg):^8s}  'advflg' = thermal  advection flag (0=T&S, 1=th&S, 2=th&T)\n"
    nmlstr += f"{str(m.advtyp):^8s}  'advtyp' = scalar   advection type (0=PCM,1=MPDATA,2=FCT2,4=FCT4)\n"
    nmlstr += f"{str(m.momtyp):^8s}  'momtyp' = momentum advection type (2=2nd order, 4=4th order)\n"
    nmlstr += f"{str(m.slip  ):^8s}  'slip  ' = +1 for free-slip, -1 for non-slip boundary conditions\n"
    nmlstr += f"{str(m.visco2):^8s}  'visco2' = deformation-dependent Laplacian  viscosity factor\n"
    nmlstr += f"{str(m.visco4):^8s}  'visco4' = deformation-dependent biharmonic viscosity factor\n"
    nmlstr += f"{str(m.facdf4):^8s}  'facdf4' =       speed-dependent biharmonic viscosity factor\n"
    nmlstr += f"{str(m.veldf2):^8s}  'veldf2' = diffusion velocity (m/s) for Laplacian  momentum dissip.\n"
    nmlstr += f"{str(m.veldf4):^8s}  'veldf4' = diffusion velocity (m/s) for biharmonic momentum dissip.\n"
    nmlstr += f"{str(m.thkdf2):^8s}  'thkdf2' = diffusion velocity (m/s) for Laplacian  thickness diffus.\n"
    nmlstr += f"{str(m.thkdf4):^8s}  'thkdf4' = diffusion velocity (m/s) for biharmonic thickness diffus.\n"
    nmlstr += f"{str(m.temdf2):^8s}  'temdf2' = diffusion velocity (m/s) for Laplacian  temp/saln diffus.\n"
    nmlstr += f"{str(m.temdfc):^8s}  'temdfc' = temp diffusion conservation (0.0,1.0 all dens,temp resp.)\n"
    nmlstr += f"{str(m.vertmx):^8s}  'vertmx' = diffusion velocity (m/s) for momentum at MICOM M.L.base\n"
    nmlstr += f"{str(m.cbar  ):^8s}  'cbar  ' = rms flow speed     (m/s) for linear bottom friction\n"
    nmlstr += f"{str(m.cb    ):^8s}  'cb    ' = coefficient of quadratic bottom friction\n"
    nmlstr += f"{str(m.drglim):^8s}  'drglim' = limiter for explicit friction (1.0 none, 0.0 implicit)\n"
    nmlstr += f"{str(m.drgscl):^8s}  'drgscl' = scale factor for tidal drag   (0.0 for no tidal drag)\n"
    nmlstr += f"{str(m.thkdrg):^8s}  'thkdrg' = thickness of bottom boundary layer for tidal drag (m)\n"
    nmlstr += f"{str(m.thkbot):^8s}  'thkbot' = thickness of bottom boundary layer (m)\n"
    nmlstr += f"{str(m.sigjmp):^8s}  'sigjmp' = minimum density jump across interfaces  (kg/m**3)\n"
    nmlstr += f"{str(m.tmljmp):^8s}  'tmljmp' = equivalent temperature jump across mixed-layer (degC)\n"
    nmlstr += f"{str(m.thkmls):^8s}  'thkmls' = reference mixed-layer thickness for SSS relaxation (m)\n"
    nmlstr += f"{str(m.thkmlt):^8s}  'thkmlt' = reference mixed-layer thickness for SST relaxation (m)\n"
    nmlstr += f"{str(m.thkriv):^8s}  'thkriv' = nominal thickness of river inflow (m)\n"
    nmlstr += f"{str(m.thkfrz):^8s}  'thkfrz' = maximum thickness of near-surface freezing zone (m)\n"
    nmlstr += f"{str(m.iceflg):^8s}  'iceflg' = sea ice model flag (0=none,1=energy loan,2=coupled/esmf)\n"
    nmlstr += f"{str(m.tfrz_0):^8s}  'tfrz_0' = ENLN: ice melting point (degC) at S=0psu\n"
    nmlstr += f"{str(m.tfrz_s):^8s}  'tfrz_s' = ENLN: gradient of ice melting point (degC/psu)\n"
    nmlstr += f"{str(m.ticegr):^8s}  'ticegr' = ENLN: temp. grad. inside ice (deg/m); =0 use surtmp\n"
    nmlstr += f"{str(m.hicemn):^8s}  'hicemn' = ENLN: minimum ice thickness (m)\n"
    nmlstr += f"{str(m.hicemx):^8s}  'hicemx' = ENLN: maximum ice thickness (m)\n"
    nmlstr += f"{str(m.ntracr):^8s}  'ntracr' = number of tracers (0=none,negative to initialize)\n"
    nmlstr += f"{str(m.trcflg):^8s}  'trcflg' = tracer flags      (one digit per tr, most sig. replicated)\n"
    nmlstr += f"{str(m.tsofrq):^8s}  'tsofrq' = number of time steps between anti-drift offset calcs\n"
    nmlstr += f"{str(m.tofset):^8s}  'tofset' = temperature anti-drift offset (degC/century)\n"
    nmlstr += f"{str(m.sofset):^8s}  'sofset' = salnity     anti-drift offset  (psu/century)\n"
    nmlstr += f"{str(m.mlflag):^8s}  'mlflag' = mixed layer flag  (0=none,1=KPP,2-3=KT,4=PWP,5=MY,6=GISS)\n"
    nmlstr += f"{str(m.pensol):^8s}  'pensol' = KT:      activate penetrating solar rad.   (0=F,1=T)\n"
    nmlstr += f"{str(m.dtrate):^8s}  'dtrate' = KT:      maximum permitted m.l. detrainment rate  (m/day)\n"
    nmlstr += f"{str(m.thkmin):^8s}  'thkmin' = KT/PWP:  minimum mixed-layer thickness (m)\n"
    nmlstr += f"{str(m.dypflg):^8s}  'dypflg' = KT/PWP:  diapycnal mixing flag (0=none, 1=KPP, 2=explicit)\n"
    nmlstr += f"{str(m.mixfrq):^8s}  'mixfrq' = KT/PWP:  number of time steps between diapycnal mix calcs\n"
    nmlstr += f"{str(m.diapyc):^8s}  'diapyc' = KT/PWP:  diapycnal diffusivity x buoyancy freq. (m**2/s**2)\n"
    nmlstr += f"{str(m.rigr  ):^8s}  'rigr  ' = PWP:     critical gradient richardson number\n"
    nmlstr += f"{str(m.ribc  ):^8s}  'ribc  ' = PWP:     critical bulk     richardson number\n"
    nmlstr += f"{str(m.rinfty):^8s}  'rinfty' = KPP:     maximum  gradient richardson number (shear inst.)\n"
    nmlstr += f"{str(m.ricr  ):^8s}  'ricr  ' = KPP:     critical bulk     richardson number\n"
    nmlstr += f"{str(m.bldmin):^8s}  'bldmin' = KPP:     minimum surface boundary layer thickness (m)\n"
    nmlstr += f"{str(m.bldmax):^8s}  'bldmax' = K-PROF:  maximum surface boundary layer thickness (m)\n"
    nmlstr += f"{str(m.cekman):^8s}  'cekman' = KPP/KT:  scale factor for Ekman depth\n"
    nmlstr += f"{str(m.cmonob):^8s}  'cmonob' = KPP:     scale factor for Monin-Obukov depth\n"
    nmlstr += f"{str(m.bblkpp):^8s}  'bblkpp' = KPP:     activate bottom boundary layer    (0=F,1=T)\n"
    nmlstr += f"{str(m.shinst):^8s}  'shinst' = KPP:     activate shear instability mixing (0=F,1=T)\n"
    nmlstr += f"{str(m.dbdiff):^8s}  'dbdiff' = KPP:     activate double diffusion  mixing (0=F,1=T)\n"
    nmlstr += f"{str(m.nonloc):^8s}  'nonloc' = KPP:     activate nonlocal b. layer mixing (0=F,1=T)\n"
    nmlstr += f"{str(m.botdiw):^8s}  'botdiw' = GISS:    activate bot.enhan.int.wav mixing (0=F,1=T)\n"
    nmlstr += f"{str(m.difout):^8s}  'difout' = K-PROF:  output visc/diff coffs in archive (0=F,1=T)\n"
    nmlstr += f"{str(m.difsmo):^8s}  'difsmo' = K-PROF:  activate horiz smooth diff coeffs (0=F,1=T)\n"
    nmlstr += f"{str(m.difm0):^8s}  'difm0 ' = KPP:     max viscosity   due to shear instability (m**2/s)\n"
    nmlstr += f"{str(m.difs0):^8s}  'difs0 ' = KPP:     max diffusivity due to shear instability (m**2/s)\n"
    nmlstr += f"{str(m.difmiw):^8s}  'difmiw' = KPP:     background/internal wave viscosity       (m**2/s)\n"
    nmlstr += f"{str(m.difsiw):^8s}  'difsiw' = KPP:     background/internal wave diffusivity     (m**2/s)\n"
    nmlstr += f"{str(m.dsfmax):^8s}  'dsfmax' = KPP:     salt fingering diffusivity factor        (m**2/s)\n"
    nmlstr += f"{str(m.rrho0):^8s}  'rrho0 ' = KPP:     salt fingering rp=(alpha*delT)/(beta*delS)\n"
    nmlstr += f"{str(m.cs):^8s}  'cs    ' = KPP:     value for nonlocal flux term\n"
    nmlstr += f"{str(m.cstar):^8s}  'cstar ' = KPP:     value for nonlocal flux term\n"
    nmlstr += f"{str(m.cv):^8s}  'cv    ' = KPP:     buoyancy frequency ratio (0.0 to use a fn. of N)\n"
    nmlstr += f"{str(m.c11):^8s}  'c11   ' = KPP:     value for turb velocity scale\n"
    nmlstr += f"{str(m.hblflg):^8s}  'hblflg' = KPP:     b. layer interp. flag (0=const.,1=linear,2=quad.)\n"
    nmlstr += f"{str(m.niter):^8s}  'niter ' = KPP:     iterations for semi-implicit soln. (2 recomended)\n"
    nmlstr += f"{str(m.fltflg):^8s}  'fltflg' = FLOATS: synthetic float flag (0=no; 1=yes)\n"
    nmlstr += f"{str(m.nfladv):^8s}  'nfladv' = FLOATS: advect every nfladv bacl. time steps (even, >=4)\n"
    nmlstr += f"{str(m.nflsam):^8s}  'nflsam' = FLOATS: output (0=every nfladv steps; >0=no. of days)\n"
    nmlstr += f"{str(m.intpfl):^8s}  'intpfl' = FLOATS: horiz. interp. (0=2nd order+n.n.; 1=n.n. only)\n"
    nmlstr += f"{str(m.iturbv):^8s}  'iturbv' = FLOATS: add horiz. turb. advection velocity (0=no; 1=yes)\n"
    nmlstr += f"{str(m.ismpfl):^8s}  'ismpfl' = FLOATS: sample water properties at float (0=no; 1=yes)\n"
    nmlstr += f"{str(m.tbvar):^8s}  'tbvar ' = FLOATS: horizontal turb. vel. variance scale (m**2/s**2)\n"
    nmlstr += f"{str(m.tdecri):^8s}  'tdecri' = FLOATS: inverse decorrelation time scale (1/day)\n"
    nmlstr += f"{str(m.lbflag):^8s}  'lbflag' = lateral barotropic bndy flag (0=none, 1=port, 2=input)\n"
    nmlstr += f"{str(m.tidflg):^8s}  'tidflg' = TIDES: tidal forcing flag    (0=none,1=open-bdy,2=bdy&body)\n"
    nmlstr += f"{str(m.tidcon):^8s}  'tidcon' = TIDES: 1 digit per (Q1K2P1N2O1K1S2M2), 0=off,1=on\n"
    nmlstr += f"{str(m.tidsal):^8s}  'tidsal' = TIDES: scalar self attraction and loading factor (<0: file)\n"
    nmlstr += f"{str(m.tidgen):^8s}  'tidgen' = TIDES: generic time (0=F,1=T)\n"
    nmlstr += f"{str(m.tidrmp):^8s}  'tidrmp' = TIDES:            ramp time (days)\n"
    nmlstr += f"{str(m.tid_t0):^8s}  'tid_t0' = TIDES: origin for ramp time (model day)\n"
    nmlstr += f"{str(m.clmflg):^8s}  'clmflg' = climatology frequency flag   (6=bimonthly, 12=monthly)\n"
    nmlstr += f"{str(m.wndflg):^8s}  'wndflg' = wind stress input flag (0=none,1=u/v-grid,2,3=p-grid)\n"
    nmlstr += f"{str(m.ustflg):^8s}  'ustflg' = ustar forcing     flag        (3=input,1,2=wndspd,4=stress)\n"
    nmlstr += f"{str(m.flxflg):^8s}  'flxflg' = thermal forcing   flag (0=none,3=net-flux,1,2,4=sst-based,99=thermf_nersc)\n"
    nmlstr += f"{str(m.empflg):^8s}  'empflg' = E-P     forcing   flag (0=none,3=net_E-P, 1,2,4=sst-bas_E)\n"
    nmlstr += f"{str(m.dswflg):^8s}  'dswflg' = diurnal shortwave flag (0=none,1=daily to diurnal corr.)\n"
    nmlstr += f"{str(m.sssflg):^8s}  'sssflg' = SSS relaxation flag (0=none,1=clim)\n"
    nmlstr += f"{str(m.lwflag):^8s}  'lwflag' = longwave (SST) flag (0=none,1=clim,2=atmos)\n"
    nmlstr += f"{str(m.sstflg):^8s}  'sstflg' = SST relaxation flag (0=none,1=clim,2=atmos,3=observed)\n"
    nmlstr += f"{str(m.icmflg):^8s}  'icmflg' = ice mask       flag (0=none,1=clim,2=atmos,3=obs/coupled)\n"
    nmlstr += f"{str(m.flxoff):^8s}  'flxoff' = net flux offset flag                 (0=F,1=T)\n"
    nmlstr += f"{str(m.flxsmo):^8s}  'flxsmo' = smooth surface fluxes                (0=F,1=T)\n"
    nmlstr += f"{str(m.relax):^8s}  'relax ' = activate lateral boundary nudging    (0=F,1=T)\n"
    nmlstr += f"{str(m.trcrlx):^8s}  'trcrlx' = activate lat. bound. tracer nudging  (0=F,1=T)\n"
    nmlstr += f"{str(m.priver):^8s}  'priver' = rivers as a precipitation bogas      (0=F,1=T)\n"
    nmlstr += f"{str(m.epmass):^8s}  'epmass' = treat evap-precip as a mass exchange (0=F,1=T)\n"
    with open(os.path.join(run_dir, 'blkdat.input'), 'wt') as f:
        f.write(nmlstr)

    refyear = time.year
    day1 = int(time.strftime('%j'))
    hour1 = int(time.strftime('%H'))
    day2 = day1 + int(forecast_period / 24)
    hour2 = (hour1 + forecast_period) % 24

    nmlstr =  f"3.1          # version of inputfile (limits.dat)            DO NOT CHANGE\n"
    nmlstr += f"TP4          # (rungen)   Version number of run\n"
    nmlstr += f"{refyear}         # Reference year for simulation  (365. for spinup)\n"
    nmlstr += f" {day1:03d} {hour1:02d}      # (nday1)    First day of integration NOTE FORMAT F9.2\n"
    nmlstr += f" {day2:03d} {hour2:02d}      # (nday2)    Last day of integration  NOTE FORMAT F9.2\n"
    nmlstr += f"{m.forcing_frc:5s} {m.forcing_clm:5s}  # forcing option, month, ecmwf, ncepr, ecmo, ecnc\n"
    nmlstr += f"{m.temp_relax_tscale:6.1f}       # temperature relaxation time scale (F in blkdat.input)\n"
    nmlstr += f"{m.saln_relax_tscale:6.1f}       # salinity    relaxation time scale\n"
    nmlstr += f"{bool_str(m.accum_avg)}  {m.accum_avg_intv_h}         # laverage n Accumulate monthly averages every n hours\n"
    nmlstr += f"T  T         # Switch on and accumulate (T) or overwrite (F) daily averages \n"
    nmlstr += f"{bool_str(m.nestoflag)}  6         # lnesto, nestdto - saves nesting bnd cond at nestdto intervals\n"
    nmlstr += f"{bool_str(m.nestiflag)}  6         # lnesti, nestdti - read and apply nesting bnd cond at nestdto intervals\n"
    nmlstr += f"{bool_str(m.tideflag)} {m.tidechoice} F F    # Tides (true,  CSR/FES, apply currents)\n"
    nmlstr += f"{bool_str(m.gpflag)}            # lgridp     Activate storage of gridpoint information\n"
    nmlstr += f"# Days     Assi  Diagno.  Restart            (1992 for first synoptic experiment)\n"
    nmlstr += f"{day1:9.2f}  F     T        T\n"
    nmlstr += f"{day2:9.2f}  F     T        T\n"
    with open(os.path.join(run_dir, 'infile.in'), 'wt') as f:
        f.write(nmlstr)

    nmlstr =  f"{str(m.fversn):^8s} 'fversn' = version of inputfile (infile2.in)\n"
    nmlstr += f"{str(m.randf):^8s} 'randf ' = 1=random forcing, 0= no random forcing\n"
    nmlstr += f"{str(m.seed):^8s} 'seed  ' = Random forcing seed\n"
    nmlstr += f"{str(m.vslp):^8s} 'vslp  ' = Variance in slp\n"
    nmlstr += f"{str(m.vtaux):^8s} 'vtaux ' = Variance in tau_x\n"
    nmlstr += f"{str(m.vtauy):^8s} 'vtauy ' = Variance in tau_y\n"
    nmlstr += f"{str(m.vwspd):^8s} 'vwspd ' = Variance in tau_y\n"
    nmlstr += f"{str(m.vcloud):^8s} 'vcloud' = Variance in clouds\n"
    nmlstr += f"{str(m.vtair):^8s} 'vtair ' = Variance in air temperature\n"
    nmlstr += f"{str(m.vprcp):^8s} 'vprcp ' = Variance in precipitation\n"
    nmlstr += f"{str(m.vrlhum):^8s} 'vrlhum' = Variance in relative humidity\n"
    nmlstr += f"{str(m.scorr):^8s} 'scorr ' = Horizontal radius of correlation (meters)\n"
    nmlstr += f"{str(m.tcorr):^8s} 'tcorr ' = Temporal radius of correlation (days)\n"
    nmlstr += f"{str(m.prsflg):^8s} 'prsflg' = 0 uncorr wind/slp, 1= wind from slp, 2=wind from slp limited by wndspd\n"
    with open(os.path.join(run_dir, 'infile2.in'), 'wt') as f:
        f.write(nmlstr)

    fdtime = dayfor(m.yrflag, refyear, day1+1, hour1)
    ldtime = dayfor(m.yrflag, refyear, day2+1, hour2)
    nmlstr = "{:14.5f} {:14.5f}".format(fdtime, ldtime)
    with open(os.path.join(run_dir, 'limits'), 'wt') as f:
        f.write(nmlstr)

    nmlstr =  f"{str(m.nports):^8s} 'nports   ' = number of ports\n"
    nmlstr += f"{str(m.pefold):^8s} 'pefold   ' = port transport e-folding time in days\n"
    for i in range(m.nports):
        nmlstr += f"{str(m.kdport[i]):^8s} 'kdport   ' = port orientation (1=N, 2=S, 3=E, 4=W)\n"
        nmlstr += f"{str(m.ifport[i]):^8s} 'ifport   ' = first i-index\n"
        nmlstr += f"{str(m.ilport[i]):^8s} 'ilport   ' = last  i-index (=ifport for east/west port)\n"
        nmlstr += f"{str(m.jfport[i]):^8s} 'jfport   ' = first j-index\n"
        nmlstr += f"{str(m.jlport[i]):^8s} 'jlport   ' = last  j-index (=jfport for north/south port)\n"
        nmlstr += f"{str(m.svpnow[i]):^8s} 'svpnow   ' = existing port transport in Sv (+ve towards E or S)\n"
        nmlstr += f"{str(m.svport[i]):^8s} 'svport   ' = target   port transport in Sv (+ve towards E or S)\n"
    with open(os.path.join(run_dir, 'ports.input'), 'wt') as f:
        f.write(nmlstr)

    nmlstr =  f"1.1      # File version\n"
    nmlstr += f"{str(m.evp_time_step):^8s} # EVP time step\n"
    nmlstr += f"{str(m.evp_ice_strength):^8s} # EVP ice strength\n"
    nmlstr += f"{str(m.evp_sic_fac):^8s} # EVP ice concentration factor\n"
    nmlstr += f"{str(m.evp_n_subcycles):^8s} # EVP nb of subcycles=elastic_tstep/dyn_tstep\n"
    with open(os.path.join(run_dir, 'infile.evp'), 'wt') as f:
        f.write(nmlstr)

    nmlstr =  f"1.1      # FILE    : Version number of this file\n"
    nmlstr += f"{str(m.albedo_melt_ice):^8s} # ALBEDO  : Albedo value of melting ice []\n"
    nmlstr += f"{str(m.albedo_dry_ice):^8s} # ALBEDO  : Albedo value of dry ice []\n"
    nmlstr += f"{str(m.albedo_snow_min):^8s} # ALBEDO  : Minimum albedo value of snow []\n"
    nmlstr += f"{str(m.albedo_snow_max):^8s} # ALBEDO  : Maximum albedo value of snow []\n"
    nmlstr += f"{str(m.frozen_ice_thick):^8s} # FREEZE  : Initial thickness of frozen ice [m]\n"
    nmlstr += f"{str(m.lead_max_sic):^8s} # LEAD    : Maximum value for ice concentration []\n"
    nmlstr += f"{str(m.snow_limit):^8s} # SNWLIM  : Maximum allowed snow thickness [m]\n"
    nmlstr += f"{str(m.qstore):^8s} # QSTORE  : Max heat store in frac. of ice latent heat []\n"
    nmlstr += f"{str(m.lateral_melt):^8s} # LATERAL MELT: 0=none, 1=standard (Drange94), 2=Hakkinen & Mellor\n"
    with open(os.path.join(run_dir, 'infile.icestate'), 'wt') as f:
        f.write(nmlstr)




import os
import glob
import numpy as np
from NEDAS.utils.conversion import dt1h
from NEDAS.models.topaz.time_format import datetojul
from NEDAS.datasets import Dataset
from .uf_data import read_uf_data

class TopazPrepObs(Dataset):
    """
    Observations already preprocessed by TOPAZ EnKF Prep_Routines, saved in unformatted binary format
    """
    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        self.variables = {
            'ocean_temp': {'name':'TEM', 'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'C'},
            'ocean_saln': {'name':'SAL', 'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'psu'},
            'ocean_surf_temp': {'name':'SST', 'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'C'},
            'ocean_surf_height_anomaly': {'name':'TSLA', 'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'m'},
            'seaice_conc': {'name':'ICEC', 'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':1},
            'seaice_thick': {'name':'HICE', 'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'m'},
            'seaice_drift': {'name':'IDRFT', 'dtype':'float', 'is_vector':True, 'z_units':'m', 'units':'km'},
            }

        self.obs_operator = {
            'seaice_drift': self.get_seaice_drift,
            }

    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        name = kwargs['name']
        time = kwargs['time']
        path = kwargs['path']

        vname = self.variables[name]['name']
        dirname = vname
        native_name = vname
        if name == 'seaice_drift':
            native_name += '?'   ##i=1, 2, 3, 4, 5: 2day drift in km with starting day = current day -i-1

        file_list = []
        if time is not None:
            time_str = "{:5d}".format(int(datetojul(time)))
        else:
            time_str = '?????'
        search = os.path.join(path, dirname, 'obs_'+native_name+'_'+time_str+'.uf')
        file_list = glob.glob(search)

        assert len(file_list)>0, 'no matching files found: '+search
        return file_list

    def random_network(self, **kwargs):
        raise NotImplementedError('random_network: ERROR: random_network not implemented for topaz dataset')

    def read_obs(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        name = kwargs['name']
        time = kwargs['time']
        grid = kwargs['grid']
        model = kwargs.get('model')
        if self.vthin:
            ##check if model.z is available for vertical thinning
            if model is None or not hasattr(model, 'z') or model.z is None:
                print("WARNING: topaz.dataset: model.z levels are not provided, setting vthin to False")
                self.vthin = False

        is_vector = self.variables[name]['is_vector']
        obs_seq = {'obs':[], 't':[], 'z':[], 'y':[], 'x':[], 'err_std':[],
                   'ipiv':[], 'jpiv':[], 'ns':[], 'in_coords':[], 'stat':[], 'i_orig_grid':[], 'j_orig_grid':[], 'h':[], 'date':[]}

        try:
            fname = self.filename(**kwargs)
        except AssertionError:
            ##just return empty obs_seq if no matching files found
            for key in obs_seq.keys():
                obs_seq[key] = np.array([])
            return obs_seq

        for file in fname:
            data = read_uf_data(file)

            if is_vector:
                nobs = len(data) // 2
            else:
                nobs = len(data)

            # if ice drift obs, get the time lag from file name
            obsType = os.path.basename(file).split('_')[1]
            if obsType[:5] == 'IDRFT':
                lag_days = int(obsType[5]) + 1
                t = time - dt1h * 24 * lag_days   ##start time of the drift vectors
            else:
                t = time

            for i in range(nobs):
                if is_vector:
                    obs = [data[i][0], data[nobs+i][0]]
                else:
                    obs = data[i][0]
                obs_seq['obs'].append(obs)
                obs_seq['t'].append(t)
                obs_seq['z'].append(-data[i][5])  ##depth
                lon, lat = data[i][3], data[i][4]
                x, y = grid.proj(lon, lat)
                obs_seq['y'].append(y)
                obs_seq['x'].append(x)
                obs_seq['err_std'].append(np.sqrt(data[i][1]))

                ##additional info
                obs_seq['ipiv'].append(data[i][6])
                obs_seq['jpiv'].append(data[i][7])
                obs_seq['ns'].append(data[i][8])
                obs_seq['in_coords'].append([data[i][9], data[i][10], data[i][11], data[i][12]])
                obs_seq['stat'].append(data[i][13])
                obs_seq['i_orig_grid'].append(data[i][14])
                obs_seq['j_orig_grid'].append(data[i][15])
                obs_seq['h'].append(data[i][16])
                obs_seq['date'].append(data[i][17])

        if self.vthin and name in ['ocean_temp', 'ocean_saln']:
            ##thin observation profiles vertically by picking those closest to model levels only
            mask = np.full(len(obs_seq['obs']), False)
            

        for key in obs_seq.keys():
            obs_seq[key] = np.array(obs_seq[key])

        if is_vector:
            obs_seq['obs'] = obs_seq['obs'].T  ##make vector dimension [2,nobs]
        return obs_seq

    def get_seaice_drift(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        grid = kwargs['grid']  ##output grid
        path = kwargs['path']
        member = kwargs['member']
        model = kwargs['model']  ##model object
        model.grid.set_destination_grid(grid)

        start_x = kwargs['x']
        start_y = kwargs['y']
        start_t = kwargs['t']
        x = start_x.copy()
        y = start_y.copy()
        obs_seq = np.zeros((2,)+x.shape)

        ##go through each of the 5 start times
        for uniq_start_t in np.unique(start_t):
            ind = np.where(start_t==uniq_start_t)

            for day in range(2):
                t = uniq_start_t + day * dt1h * 24

                ##get daily mean seaice velocity
                model_vel = model.read_var(path=path, name='seaice_velocity_daily', time=t, member=member, units='km/day')
                ##convert to grid
                vel = model.grid.convert(model_vel, is_vector=True)
                ##get velocity at x, y position
                u = grid.interp(vel[0,...], x[ind], y[ind])  ##km over 1 day
                v = grid.interp(vel[1,...], x[ind], y[ind])

                ##increment the x,y components
                x[ind] += u * 1000.
                y[ind] += v * 1000.

        obs_seq[0, :] = (x - start_x) / 1000.
        obs_seq[1, :] = (y - start_y) / 1000.

        return obs_seq




import os
import inspect
from abc import ABC, abstractmethod
import numpy as np
from datetime import datetime
from NEDAS.config import parse_config

class Dataset(ABC):
    """
    Dataset class (template for specific dataset sources)
    """
    def  __init__(self, config_file=None, parse_args=False, **kwargs):

        ##parse config file and obtain a list of attributes
        code_dir = os.path.dirname(inspect.getfile(self.__class__))
        config_dict = parse_config(code_dir, config_file, parse_args, **kwargs)
        for key, value in config_dict.items():
            setattr(self, key, value)

        if not hasattr(self, 'dataset_dir'):
            self.dataset_dir = '.'

        if not hasattr(self, 'obs_window_min'):
            self.obs_window_min = 0
        if not hasattr(self, 'obs_window_max'):
            self.obs_window_max = 0

        self.variables = {}

    def parse_kwargs(self, **kwargs):
        """
        Parse the input kwargs to pinpoint a specific file/variable...
        """
        ##args to pinpoint a certain observatino record, used by read_obs, etc.
        if 'path' not in kwargs:
            kwargs['path'] = self.dataset_dir

        if 'name' not in kwargs:
            kwargs['name'] = list(self.variables.keys())[0]
        assert kwargs['name'] in self.variables, f"'{kwargs['name']}' is not defined in dataset variables"

        if 'time' not in kwargs:
            kwargs['time'] = None
        if kwargs['time'] is not None:
            assert isinstance(kwargs['time'], datetime), "kwargs 'time' is not a datetime object"

        if 'units' not in kwargs:
            kwargs['units'] = self.variables[kwargs['name']]['units']

        ##other args, set default values if not specified
        for key in ['model', 'grid', 'mask']:
            if key not in kwargs:
                kwargs[key] = None
        for key in ['nobs', 'obs_window_min', 'obs_window_max']:
            if key not in kwargs:
                kwargs[key] = 0

        return kwargs

    @abstractmethod
    def read_obs(self, **kwargs) -> dict:
        """
        Return observation sequence matching the given kwargs
        """
        obs_seq = {'obs':[], 't':[], 'z':[], 'y':[], 'x':[], 'err_std':[], }
        return obs_seq



import numpy as np
from NEDAS.utils.njit import njit

@njit
def gaspari_cohn_func(dist, roi):
    """
    Gaspari-Cohn localization function.

    Args:
        dist (np.ndarray): Distance between observation and state (being updated)
        roi (float or np.ndarray): The radius of influence, distance beyond which local_factor is tapered to 0

    Returns:
        np.ndarray: The localization factor, same shape as dist
    """
    shape = dist.shape
    dist = dist.flatten()
    lfactor = np.zeros(dist.shape)
    r = dist / (roi / 2)

    ind1 = np.where(r<1)
    loc1 = (((-0.25*r + 0.5)*r + 0.625)*r - 5.0/3.0) * r**2 + 1
    lfactor[ind1] = loc1[ind1]

    ind2 = np.where(np.logical_and(r>=1, r<2))
    r[np.where(r==0)] = 1e-10  ##avoid divide by 0
    loc2 = ((((r/12.0 - 0.5)*r + 0.625)*r + 5.0/3.0)*r - 5.0)*r + 4 - 2.0/(3.0*r)
    lfactor[ind2] = loc2[ind2]

    return lfactor.reshape(shape)

@njit
def step_func(dist, roi):
    """
    Step function for localization.
    """
    shape = dist.shape
    dist = dist.flatten()
    lfactor = np.zeros(dist.shape)
    lfactor[np.where(dist<=roi)] = 1.0
    return lfactor.reshape(shape)

@njit
def exponential_func(dist, roi):
    """
    Exponential decay localization function.
    """
    shape = dist.shape
    dist = dist.flatten()
    lfactor = np.exp(-dist/roi)
    return lfactor.reshape(shape)



###Utility function for handling qgmodel fields
import numpy as np
from numpy.fft import fft2, ifft2, fftshift, ifftshift

##file i/o
def read_data_bin(filename, kmax, nz, k, r=0):
    nkx = 2*kmax+1
    nky = kmax+1
    fieldk = np.zeros((nkx, nky), dtype=complex)
    with open(filename, 'rb') as f:
        for i in range(nky):
            f.seek(((2*r*nz+k)*nky+i)*nkx*8)
            fieldk[:, i].real = np.fromfile(f, dtype=float, count=nkx)
            f.seek(((2*r*nz+nz+k)*nky+i)*nkx*8)
            fieldk[:, i].imag = np.fromfile(f, dtype=float, count=nkx)
        return fieldk


def write_data_bin(filename, fieldk, kmax, nz, k, r=0):
    ##note: filename should exists before calling this to write content to it
    nkx = 2*kmax+1
    nky = kmax+1
    assert fieldk.shape == (nkx, nky), 'input fieldk shape mismatch with kmax'
    with open(filename, 'r+b') as f:
        for i in range(nky):
            f.seek(((2*r*nz+k)*nky+i)*nkx*8)
            f.write(fieldk[:, i].real.tobytes())
            f.seek(((2*r*nz+nz+k)*nky+i)*nkx*8)
            f.write(fieldk[:, i].imag.tobytes())


##conversion between physical and spectral spaces
def fullspec(hfield):
    nkx, nky = hfield.shape
    kmax = nky-1
    hres = nkx+1
    sfield = np.zeros((hres, hres), dtype=complex)
    fup = hfield
    fup[kmax-1::-1, 0] = np.conjugate(fup[kmax+1:nkx, 0])
    fdn = np.conjugate(fup[::-1, nky-1:0:-1])
    sfield[1:hres, nky:hres] = fup
    sfield[1:hres, 1:nky] = fdn
    return sfield


def halfspec(sfield):
    n1, n2 = sfield.shape
    kmax = int(n1/2)-1
    hfield = sfield[1:n1, kmax+1:n2]
    return hfield


def spec2grid(fieldk):
    nkx, nky = fieldk.shape
    nx = nkx+1
    ny = 2*nky
    fieldk = ifftshift(fullspec(fieldk))
    field = nx*ny*np.real(ifft2(fieldk))
    return field


def grid2spec(field):
    nx, ny = field.shape
    nkx = nx-1
    nky = int(ny/2)
    fieldk = fft2(field)/nx/ny
    fieldk = halfspec(fftshift(fieldk))
    return fieldk


##conversion between variables in spectral space
def get_wn(psik):
    nkx, nky = psik.shape
    kmax = nky-1
    kx_, ky_ = np.mgrid[-kmax:kmax+1, 0:kmax+1]
    return kx_, ky_


def psi2zeta(psik):
    kx_, ky_ = get_wn(psik)
    zetak = -(kx_**2 + ky_**2) * psik
    return zetak


def psi2temp(psik):
    kx_, ky_ = get_wn(psik)
    tempk = -np.sqrt(kx_**2 + ky_**2) * psik
    return tempk


def psi2u(psik):
    kx_, ky_ = get_wn(psik)
    uk = -1j * ky_ * psik
    return uk


def psi2v(psik):
    kx_, ky_ = get_wn(psik)
    vk = 1j * kx_ * psik
    return vk


def uv2zeta(uk, vk):
    kx_, ky_ = get_wn(uk)
    zetak = 1j*kx_*vk - 1j*ky_*uk
    return zetak


def zeta2psi(zetak):
    nkx, nky = zetak.shape
    kmax = nky-1
    kx_, ky_ = get_wn(zetak)
    k2_ = kx_**2 + ky_**2
    k2_[kmax, 0] = 1  #set irrelavent point to 1 to avoid singularity in inversion
    psik = -(1.0/k2_) * zetak
    return psik


def temp2psi(tempk):
    nkx, nky = tempk.shape
    kmax = nky-1
    kx_, ky_ = get_wn(tempk)
    k1_ = np.sqrt(kx_**2 + ky_**2)
    k1_[kmax, 0] = 1
    psik = -(1.0/k1_) * tempk
    return psik




import os
import stat
import subprocess
import tempfile
from time import sleep
from NEDAS.utils.conversion import seconds_to_timestr
from NEDAS.job_submitters.base import JobSubmitter

class OARJobSubmitter(JobSubmitter):
    """JobSubmitter Class customized for OAR schedulers"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        ##temporary node file
        if not self.run_separate_jobs:
            with tempfile.NamedTemporaryFile(delete=False, dir=self.run_dir) as file:
                self.node_file = file.name

        ##host specific settings
        if self.host == 'gricad':
            p = subprocess.run("hostname", capture_output=True, text=True)
            if p.stdout.replace('\n', '').replace(' ', '') in ['dahu-oar3', 'f-dahu']:
                self.job_submit_node = None  ##don't need ssh for oarsub on compute nodes
            else:
                ##job submit node based on queue type
                if self.queue == 'devel':
                    self.job_submit_node = 'dahu-oar3'
                else:
                    self.job_submit_node = 'f-dahu'

    @property
    def nproc_avail(self):
        return self.nnode_avail * self.ppn

    @property
    def node_list_avail(self):
        try:
            with open(os.environ['OAR_NODE_FILE'], 'r') as node_file:
                node_list = [node for node in node_file]
        except Exception as e:
            print("ERROR while reading available compute nodes from OAR_NODE_FILE")
            raise e
        return node_list

    @property
    def nnode_avail(self):
        return len(self.node_list_avail)

    @property
    def ppn_avail(self):
        ##TODO: how to obtain this info from OAR system?
        return 32

    def update_node_file(self):
        with open(self.node_file, 'w') as file:
            for i in range(self.nnode):
                node = self.node_list_avail[self.offset_node+i]
                file.write(node+'\n')

    @property
    def execute_command(self):
        if self.run_separate_jobs:
            node_file = '$OAR_NODE_FILE'
        else:
            self.update_node_file()
            node_file = self.node_file
        return f"mpirun -np {self.nproc} --machinefile {node_file}"

    @property
    def job_array_index_name(self):
        return '$OAR_ARRAY_INDEX'

    def run_job_as_step(self, commands):
        super().run_job_as_step(commands)
        ##clean up
        os.remove(self.node_file)

    def submit_job_and_monitor(self, commands):
        ##build the job script
        with tempfile.NamedTemporaryFile(mode='w+', delete=False,
                                         dir=self.run_dir,
                                         prefix=self.job_name+'.',
                                         suffix='.sh') as job_script:
            job_script.write("#!/bin/bash\n")

            ##OAR headers
            job_script.write(f"#OAR -n {self.job_name}\n")
            job_script.write(f"#OAR -l /nodes={self.nnode}/core={self.ppn},walltime={seconds_to_timestr(self.walltime)}\n")
            job_script.write(f"#OAR --project {self.project}\n")
            if self.queue == 'devel':
                job_script.write(f"#OAR -t {self.queue}\n")

            log_file = os.path.join(self.run_dir, f"{self.job_name}.%jobid%.stdout")
            job_script.write(f"#OAR --stdout {log_file}\n")
            job_script.write(f"#OAR --stderr {log_file}\n")

            if self.use_job_array:
                job_script.write(f"#OAR --array {self.array_size}\n")

            commands = super().parse_commands(commands)
            job_script.write(commands)
            job_script.write('\n')

            self.job_script = job_script.name

        # the job script has to be executables in OAR system
        st = os.stat(self.job_script)
        os.chmod(self.job_script, st.st_mode | stat.S_IEXEC)

        ##submit the job script
        if self.job_submit_node:
            submit_cmd = ['ssh', self.job_submit_node,
                          f"oarsub -S {self.job_script}"]
        else:
            submit_cmd = ["oarsub", "-S", f"{self.job_script}"]
        process = subprocess.run(submit_cmd, capture_output=True)
        s = process.stderr.decode('utf-8')
        print(s, flush=True)
        s = process.stdout.decode('utf-8')
        print(s, flush=True)

        ##monitor the queue for job completion
        if self.use_job_array:
            self.job_id = int(s.replace(' ', '').replace('\n', '').split('OAR_ARRAY_ID=')[-1])
            while True:
                sleep(self.check_dt)
                p = subprocess.run(['ssh', self.job_submit_node,
                                    'oarstat', '-f', f'--array {self.job_id}',
                                    '| grep "state = "'], capture_output=True)
                s = p.stdout.decode('utf-8').replace(' ', '').split('\n')[:-1]
                print(s, flush=True)
                s = [string for string in s if 'state=' in string]
                print(s, flush=True)
                jobs_status = [job.split('state=')[-1].replace(' ', '') for job in s]
                # end this loop if all jobs are terminated
                if all([status == 'Terminated' for status in jobs_status]):
                    break
                if all([status == 'Error' for status in jobs_status]):
                    raise RuntimeError(f"Error job array {self.job_id}")

        else:
            self.job_id = int(s.replace(' ', '').replace('\n', '').split('OAR_JOB_ID=')[-1])
            while True:
                sleep(20)
                p = subprocess.run(['ssh', self.job_submit_node,
                                    'oarstat', '-f', f'--job {self.job_id}',
                                    '| grep "state = "'], capture_output=True)
                s = p.stdout.decode('utf-8').replace(' ', '').split('\n')[:-1]
                s = s[0]
                print(s, flush=True)
                jobs_status = s.split('state=')[-1].replace(' ', '')
                # end this loop if all jobs are terminated
                if jobs_status == 'Terminated':
                    break
                if jobs_status == 'Error':
                    raise RuntimeError(f"Error in job {self.job_id}")



import numpy as np
from NEDAS.config import Config
from .base import Transform

class Identity(Transform):
    """
    Subclass for the identity transform.
    """
    def forward_state(self, c: Config, rec: dict, field: np.ndarray) -> np.ndarray:
        return field

    def backward_state(self, c: Config, rec: dict, field: np.ndarray) -> np.ndarray:
        return field

    def forward_obs(self, c: Config, obs_rec: dict, obs_seq: dict) -> dict:
        return obs_seq

    def backward_obs(self, c: Config, obs_rec: dict, obs_seq: dict) -> dict:
        return obs_seq




import os
import sys
import tempfile
import importlib.util
import subprocess
import numpy as np
from NEDAS.utils.conversion import ensure_list, dt1h
from NEDAS.utils.progress import timer, progress_bar
from NEDAS.utils.shell_utils import makedir, run_command, run_job
from NEDAS.utils.parallel import Scheduler, bcast_by_root, distribute_tasks
from NEDAS.utils.random_perturb import random_perturb
from NEDAS import assim_tools

class OfflineFilterAnalysisScheme:
    """
    Offline filtering analysis scheme class.

    This scheme runs the 4D analysis by cycling through time steps. Running the ensemble forecast first,
    pause the model and write model states (the `prior`) to restart files at a certain time step (`analysis cycle`),
    then perform data assimilation at the analysis cycle with observations within a time window,
    finally using the updated model states (the `posterior`) as new initial conditions, the ensemble forecast
    is run again to reach the next analysis cycle, until the end of the period of interest. The length of
    forecasts between cycles is called the `cycling period`.
    """
    def __call__(self, c) -> None:
        c.show_summary()

        if c.step:
            ##if --step=STEP is specified at runtime, just run STEP and quit
            if c.step in ['perturb', 'filter', 'diagnose']:
                mpi = True
            else:
                mpi = False
            self.run_step(c, c.step, mpi)
            return

        print("Cycling start...", flush=True)
        while c.time < c.time_end:
            print(f"\n\033[1;33mCURRENT CYCLE\033[0m: {c.time} => {c.next_time}", flush=True)

            os.system("mkdir -p "+c.cycle_dir(c.time))

            if c.run_preproc:
                self.run_step(c, 'preprocess', mpi=False)
                self.run_step(c, 'perturb', mpi=True)

            ##assimilation step
            if c.run_analysis and c.time >= c.time_analysis_start and c.time <= c.time_analysis_end:
                self.run_step(c, 'filter', mpi=True)
                self.run_step(c, 'postprocess', mpi=False)

            ##advance model state to next analysis cycle
            if c.run_forecast:
                self.run_step(c, 'ensemble_forecast', mpi=False)

            ##compute diagnostics
            if c.run_diagnose:
                self.run_step(c, 'diagnose', mpi=True)

            ##advance to next cycle
            c.time = c.next_time

        print("Cycling complete.", flush=True)

    def filter(self, c):
        """
        Main method for performing the analysis step

        Args:
            c (Config): runtime configuration object
        """
        self.validate_mpi_environment(c)

        ##multiscale approach: loop over scale components and perform assimilation on each scale
        ##more complex outer loops can be implemented here
        analysis_grid = c.grid
        for c.iter in range(c.niter):
            c.print_1p(f"Running analysis for outer iteration step {c.iter}:")

            self.init_analysis_dir(c)
            c.grid = analysis_grid.change_resolution_level(c.resolution_level[c.iter])
            c.transform_funcs = assim_tools.transforms.get_transform_funcs(c)
            c.localization_funcs = assim_tools.localization.get_localization_funcs(c)
            c.inflation_func = assim_tools.inflation.get_inflation_func(c)

            state = assim_tools.state.get_state(c)
            timer(c)(state.prepare_state)(c)

            obs = assim_tools.obs.get_obs(c, state)
            timer(c)(obs.prepare_obs)(c, state)
            timer(c)(obs.prepare_obs_from_state)(c, state, 'prior')

            assimilator = assim_tools.assimilators.get_assimilator(c)
            timer(c)(assimilator.assimilate)(c, state, obs)

            updator = assim_tools.updators.get_updator(c)
            timer(c)(updator.update)(c, state)

    def ensemble_forecast(self, c):
        """
        Ensemble forecast step.

        """
        for model_name, model in c.models.items():
            path = c.forecast_dir(c.time, model_name)
            makedir(path)
            print(f"Running {model_name} ensemble forecast:", flush=True)

            if model.ens_run_type == 'batch':
                opts = self.get_task_opts(c, path=path, nens=c.nens)
                model.run_batch(**opts)

            elif model.ens_run_type == 'scheduler':
                opts = self.get_task_opts(c, path=path)
                walltime = getattr(model, 'walltime', None)
                self.run_ensemble_tasks_in_scheduler(c, f'forecast_{model_name}', model.run, opts, model.nproc_per_run, walltime)

            else:
                raise ValueError(f"Unknown ensemble run type {model.ens_run_type} for {model_name}")

    def preprocess(self, c):
        """
        Pre-processing step before the forecast.

        This step prepares the necessary files (static data, boundary condition, namelist parameters etc.).
        Restart files from the previous step (during cycling) or from ``ens_init_dir`` (at first cycle)
        will be used as initial conditions.

        The ``preprocess`` method implemented in each model class details this step.
        """
        for model_name, model in c.models.items():
            path = c.forecast_dir(c.time, model_name)
            makedir(path)
            print(f"Preprocessing {model_name} state:", flush=True)
            restart_dir = self.get_restart_dir(c, model_name)
            opts = self.get_task_opts(c, path=path, restart_dir=restart_dir)
            self.run_ensemble_tasks_in_scheduler(c, f'preproc_{model_name}', model.preprocess, opts, model.nproc_per_util)

    def postprocess(self, c):
        """
        Post-processing step after the assimilation and before the next forecast.

        This step takes posterior model states after data assimilation. Run post-processing algorithm to remove any
        non-physical values and ensure the next model forecasts will run correctly.

        The ``postprocess`` method implemented in each model class details this step.
        """
        for model_name, model in c.models.items():
            path = c.forecast_dir(c.time, model_name)
            makedir(path)
            print(f"Postprocessing {model_name} state:", flush=True)
            restart_dir = self.get_restart_dir(c, model_name)
            opts = self.get_task_opts(c, path=path, restart_dir=restart_dir)
            self.run_ensemble_tasks_in_scheduler(c, f'postproc_{model_name}', model.postprocess, opts, model.nproc_per_util)

    def perturb(self, c):
        """
        Perturbation step.

        This step adds random perturbations to the model initial and/or boundary conditions,
        at the first or all the analysis cycles.

        The ``perturb`` section in configuration file defines the scheme and parameters for the perturbation.
        The `utils.random_perturb`` module implements the random field generator functions.
        """
        if c.perturb is None:
            c.print_1p(f"No perturbation defined in config, exiting.\n")
            return
        c.print_1p(f"Perturbing state:")

        ##clean perturb files in current cycle dir
        for rec in c.perturb:
            perturb_dir = os.path.join(c.forecast_dir(c.time, rec['model_src']), 'perturb')
            if c.pid==0:
                run_command(f"rm -rf {perturb_dir}; mkdir -p {perturb_dir}")
        c.comm.Barrier()

        ##distribute perturbation items among MPI ranks
        task_list = bcast_by_root(c.comm)(self.distribute_perturb_tasks)(c)

        c.pid_show = [p for p,lst in task_list.items() if len(lst)>0][0]

        ##first go through the fields to count how many (for showing progress)
        nfld = 0
        for rec in task_list[c.pid]:
            model_name = rec['model_src']
            model = c.models[model_name]
            vname = ensure_list(rec['variable'])[0]
            dt = model.variables[vname]['dt']
            niter = c.cycle_period // dt + 1
            for n in range(niter):
                for k in model.variables[vname]['levels']:
                    nfld += 1

        ##actually go through the fields to perturb now
        fld_id = 0
        for rec in task_list[c.pid]:
            model_name = rec['model_src']
            model = c.models[model_name]  ##model class object
            mem_id = rec['member']
            mstr = f'_mem{mem_id+1:03d}'
            path = c.forecast_dir(c.time, model_name)
            variable_list = ensure_list(rec['variable'])

            ##check if previous perturb is available from past cycles
            perturb = {}
            for vname in variable_list:
                psfile = os.path.join(c.forecast_dir(c.prev_time, model_name), 'perturb', vname+mstr+'.npy')
                if os.path.exists(psfile):
                    perturb[vname] = np.load(psfile)
                else:
                    perturb[vname] = None

            # get number of time steps for this set of variables
            # perturbation will be generated for all time steps if variable is available
            dt = max([model.variables[v]['dt'] for v in variable_list])
            nstep = c.cycle_period // dt + 1
            for n in range(nstep):
                t = c.time + n * dt * dt1h

                # TODO: perturbation for each k level is drawn independently, can be improved
                # by introducing a vertical correlation length scale, or using EOF modes.
                # Note: assuming all variables in the list have the same k levels
                for k in model.variables[variable_list[0]]['levels']:
                    fld_id += 1
                    if c.debug:
                        print(f"PID {c.pid:4}: perturbing mem{mem_id+1:03} {variable_list} at {t} level {k}", flush=True)
                    else:
                        c.print_1p(progress_bar(fld_id, nfld+1))

                    vname =variable_list[0]  ##note: all variables in the list shall have same dt and k levels
                    model.read_grid(path=path, name=vname, time=t, member=mem_id, k=k)
                    model.grid.set_destination_grid(c.grid)
                    c.grid.set_destination_grid(model.grid)

                    # collect variable fields
                    fields = {}
                    for vname in variable_list:
                        ##read variable from model state
                        fld = model.read_var(path=path, name=vname, time=t, member=mem_id, k=k)
                        ##convert to analysis grid
                        fields[vname] = model.grid.convert(fld, is_vector=model.variables[vname]['is_vector'])

                    ##generate perturbation on analysis grid
                    fields_pert, perturb = random_perturb(c.grid, fields, prev_perturb=perturb, dt=dt, n=n, **rec)

                    if rec['type'].split(',')[0]=='displace' and hasattr(model, 'displace'):
                        ##use model internal method to apply displacement perturbations directly
                        model.displace(perturb, path=path, time=t, member=mem_id, k=k)
                    else:
                        ##convert from analysis grid to model grid, and
                        ##write the perturbed variables back to model state files
                        for vname in variable_list:
                            fld = c.grid.convert(fields_pert[vname], is_vector=model.variables[vname]['is_vector'])
                            model.write_var(fld, path=path, name=vname, time=t, member=mem_id, k=k)

            ##save a copy of perturbation at next_t, for use by next cycle
            for vname in variable_list:
                psfile = os.path.join(path, 'perturb', vname+mstr+'.npy')
                run_command(f"mkdir -p {os.path.dirname(psfile)}")
                np.save(psfile, perturb[vname])

        c.comm.Barrier()
        c.print_1p(' done.\n')

    def diagnose(self, c):
        """
        Diagnostics step.

        This step runs diagnostics for the current analysis cycle.

        The ``diag`` section in configuration file defines the methods and parameters of the diagnostics,
        corresponding to the ``diag.method`` module that implements the particular diagnostic method.
        """
        c.print_1p(f"Running diagnostics:")

        ##get task list for each rank
        task_list = bcast_by_root(c.comm)(self.distribute_diag_tasks)(c)

        ##the processor with most work load will show progress messages
        c.pid_show = [p for p,lst in task_list.items() if len(lst)>0][0]

        ##init file locks for collective i/o
        self.init_file_locks(c)

        ntask = len(task_list[c.pid])
        for task_id, rec in enumerate(task_list[c.pid]):
            if c.debug:
                print(f"PID {c.pid:4} running diagnostics '{rec['method']}'", flush=True)
            else:
                c.print_1p(progress_bar(task_id, ntask))

            method_name = f"NEDAS.diag.{rec['method']}"
            mod = importlib.import_module(method_name)

            ##perform the diag task
            mod.run(c, **rec)

        c.comm.Barrier()
        c.print_1p(' done.\n')
        c.comm.cleanup_file_locks()

    def run_step(self, c, step, mpi):
        """
        Helper function to run this script (``offline_filter.py``) from an external call.

        Args:
            c (Config): Configuration.
            step (str): Step to run.
            mpi (bool): Whether to run the step within mpi environment.

        This is useful when each step uses different strategies in parallelization.
        For the ``filter`` , ``diagnose`` and ``perturb`` steps, an mpirun call starts the program with mpi4py support,
        while the ``ensemble_forecast``, ``preprocess`` and ``postprocess`` steps start as a serial call to python,
        and utilize a custom scheduler implemented in ``utils.parallel.Scheduler`` to spawn the tasks to multiple processors.
        """
        script_file = os.path.abspath(__file__)

        # create a temporary config yaml file to hold c, and pass into program through runtime arg
        with tempfile.NamedTemporaryFile(dir=c.work_dir,
                                         prefix='config-',
                                         suffix='.yml') as tmp_config_file:
            c.dump_yaml(tmp_config_file.name)

            print(f"\n\033[1;33mRUNNING\033[0m {step} step")
            if c.debug:
                print(f"config file: {tmp_config_file.name}")

            ##build run commands for the ensemble forecast script
            commands = ""
            if c.python_env:
                commands = f". {c.python_env}; "
            if mpi:
                if importlib.util.find_spec("mpi4py") is not None:
                    commands += f"JOB_EXECUTE {sys.executable} {script_file} -c {tmp_config_file.name}"
                else:
                    print("Warning: mpi4py is not found, will try to run with nproc=1.", flush=True)
                    commands += f"{sys.executable} {script_file} -c {tmp_config_file.name} --nproc=1"
            else:
                commands += f"{sys.executable} {script_file} -c {tmp_config_file.name}"
            commands += f" --step {step}"

            if mpi:
                job_opts = {
                    'job_name': step,
                    'run_dir': c.cycle_dir(c.time),
                    'nproc': c.nproc,
                    'debug': c.debug,
                    **(c.job_submit or {}),
                    }
                run_job(commands, **job_opts)
            else:
                p = subprocess.Popen(commands, shell=True, text=True)
                p.wait()
                if p.returncode != 0:
                    print(f"OfflineFilter: run_step '{step}' exited with error")
                    sys.exit(1)

    def get_task_opts(self, c, **other_opts):
        """
        
        """
        opts = {
            'time': c.time,
            'forecast_period': c.cycle_period,
            'time_start': c.time_start,
            'time_end': c.time_end,
            'debug': c.debug,
            **(c.job_submit or {}),
            **other_opts,
            }
        return opts

    def get_restart_dir(self, c, model_name):
        model = c.models[model_name]
        if c.time == c.time_start:
            restart_dir = model.ens_init_dir
        else:
            restart_dir = c.forecast_dir(c.prev_time, model_name)
        print(f"using restart files in {restart_dir}", flush=True)
        return restart_dir

    def run_ensemble_tasks_in_scheduler(self, c, name, func, opts, nproc_per_run, walltime=None):

        ##get number of workers to initialize the scheduler
        if c.job_submit and c.job_submit.get('run_separate_jobs', False):
            ##all jobs will be submitted to external scheduler's queue
            ##just assign a worker to each ensemble member
            nworker = min(c.nens, c.nproc)
        else:
            ##Scheduler will use nworkers to spawn ensemble member runs to
            ##the available nproc processors
            nworker = c.nproc // nproc_per_run
        scheduler = Scheduler(nworker, walltime, debug=c.debug)

        for mem_id in range(c.nens):
            scheduler.submit_job(name+f"_mem{mem_id+1:03}", func, member=mem_id, **opts)

        scheduler.start_queue() ##start the job queue
        scheduler.shutdown()
        print(' done.', flush=True)

    def distribute_perturb_tasks(self, c):
        task_list_full = []
        for perturb_rec in ensure_list(c.perturb):
            for mem_id in range(c.nens):
                task_list_full.append({**perturb_rec, 'member':mem_id})
        task_list = distribute_tasks(c.comm, task_list_full)
        return task_list

    def distribute_diag_tasks(self, c):
        """Build the full task list and distribute among mpi ranks"""
        task_list_full = []
        for rec in ensure_list(c.diag):
            ##load the module for the given method
            method_name = f"NEDAS.diag.{rec['method']}"
            module = importlib.import_module(method_name)
            ##module returns a list of tasks to be done by each processor
            if not hasattr(module, 'get_task_list'):
                task_list_full.append(rec)
                continue
            task_list_rec = module.get_task_list(c, **rec)
            for task in task_list_rec:
                task_list_full.append(task)
        ##collected full list of tasks is evenly distributed across the mpi communicator
        task_list = distribute_tasks(c.comm, task_list_full)
        return task_list

    def init_file_locks(self, c):
        """Build the full task list for the diagnostics part of the config"""
        for rec in ensure_list(c.diag):
            ##load the module for the given method
            method_name = f"NEDAS.diag.{rec['method']}"
            module = importlib.import_module(method_name)
            ##module get_file_list returns a list of files for collective i/o
            if not hasattr(module, 'get_file_list'):
                continue
            files = module.get_file_list(c, **rec)
            for file in files:
                ##create the file lock across mpi ranks for this file
                c.comm.init_file_lock(file)

    def validate_mpi_environment(self, c):
        """
        Validate the MPI environment and ensure the number of processes is consistent.
        """
        nproc = c.nproc
        nproc_actual = c.comm.Get_size()
        if nproc != nproc_actual:
            raise RuntimeError(f"Error: nproc {nproc} != mpi size {nproc_actual}")

    def init_analysis_dir(self, c):
        """
        Initialize the analysis directory.
        """
        self._analysis_dir = c.analysis_dir(c.time, c.iter)
        if c.pid == 0:
            makedir(self._analysis_dir)
            print(f"\nRunning assimilation step in {self._analysis_dir}\n", flush=True)

if __name__ == '__main__':
    # get config from runtime args, including the step to run (from --step)
    from NEDAS.config import Config
    c = Config(parse_args=True)

    if not c.step:
        raise RuntimeError("no step specified for offline filter scheme")

    scheme = OfflineFilterAnalysisScheme()
    step = getattr(scheme, c.step)
    timer(c)(step)(c)

    c.comm.finalize()



import os
import numpy as np
from NEDAS.grid import Grid1D
from NEDAS.utils.conversion import dt1h
from NEDAS.utils.shell_utils import run_command, makedir
from NEDAS.utils.netcdf_lib import nc_read_var, nc_write_var
from .core import M_nl
from NEDAS.models import Model

class Lorenz96Model(Model):
    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        self.grid = Grid1D.regular_grid(0, self.nx, 1, cyclic=True)
        self.grid.mask = np.full(self.grid.x.shape, False)

        self.variables = {'state': {'name':'state', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'*'}, }
        self.z_units = '*'

        self.run_process = None
        self.run_status = 'pending'

    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)

        if kwargs['member'] is not None:
            mstr = '_mem{:04d}'.format(kwargs['member']+1)
        else:
            mstr = ''

        assert kwargs['time'] is not None, 'missing time in kwargs'
        tstr = kwargs['time'].strftime('%Y%m%d_%H')

        return os.path.join(kwargs['path'], tstr+mstr+'.nc')

    def read_grid(self, **kwargs):
        pass

    def read_mask(self, **kwargs):
        pass

    def read_var(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)
        name = kwargs['name']
        var = nc_read_var(fname, self.variables[name]['name'])[0, ...]
        return var

    def write_var(self, var, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)
        name = kwargs['name']
        nc_write_var(fname, {'t':None, 'x':self.nx}, self.variables[name]['name'], var, recno={'t':0})

    def z_coords(self, **kwargs):
        return np.zeros(self.nx)

    def generate_initial_condition(self):
        state = np.random.normal(0, 1, self.nx)
        return state

    def preprocess(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        makedir(kwargs['path'])
        file1 = self.filename(**{**kwargs, 'path':kwargs['restart_dir']})
        file2 = self.filename(**kwargs)
        run_command(f"cp -fL {file1} {file2}")

    def postprocess(self, task_id=0, **kwargs):
        pass

    def run(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        self.run_status = 'running'

        state = self.read_var(**kwargs)

        makedir(kwargs['path'])

        next_time = kwargs['time'] + kwargs['forecast_period'] * dt1h

        next_state = M_nl(state, self.F, kwargs['forecast_period']/24, self.dt)

        self.write_var(next_state, **{**kwargs, 'time':next_time})

        self.run_status = 'complete'




import numpy as np
import pyproj
from NEDAS.grid import Grid
from NEDAS.datasets import Dataset

class Cs2SmosObs(Dataset):
    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        self.variables = {'seaice_thick': {'dtype':'float', 'is_vector':False, 'z_units':'m', 'units':'m'}, }

        proj = pyproj.Proj(self.proj)
        x, y = np.meshgrid(np.arange(self.xstart, self.xend, self.dx), np.arange(self.ystart, self.yend, self.dy))
        self.grid = Grid(proj, x, y)

    def filename(self, **kwargs):
        pass

    def read_obs(self, **kwargs):
        raise NotImplementedError
        ##TODO: implement real CS2SMOS dataset
        kwargs = super().parse_kwargs(**kwargs)
        grid = kwargs['grid']
        mask = kwargs['mask']

        obs_seq = {'obs':[], 't':[], 'z':[], 'y':[], 'x':[], 'err_std':[], }

        return obs_seq

    def random_network(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)

        if kwargs['nobs'] is None:
            nobs = 1000
        else:
            nobs = kwargs['nobs']

        grid = kwargs['grid']

        obs_x = []
        obs_y = []
        while len(obs_x) < nobs:
            y = np.random.uniform(grid.ymin, grid.ymax)
            x = np.random.uniform(grid.xmin, grid.xmax)
            valid = grid.interp(grid.mask.astype(int), x, y)
            if valid[0] == 0:
                obs_x.append(x)
                obs_y.append(y)

        obs_x = np.array(obs_x)
        obs_y = np.array(obs_y)
        obs_z = np.zeros(nobs)

        obs_seq = {'obs': np.full(nobs, np.nan),
                't': np.full(nobs, kwargs['time']),
                'z': obs_z,
                'y': obs_y,
                'x': obs_x,
                'err_std': np.ones(nobs) * kwargs['err']['std']
                }
        return obs_seq



import numpy as np
from NEDAS.utils.njit import njit
from NEDAS.assim_tools.assimilators.serial import SerialAssimilator

class EAKFAssimilator(SerialAssimilator):
    def obs_increment(self, obs_prior, obs, obs_err):
        return obs_increment_eakf(obs_prior, obs, obs_err)

    def update_local_state(self, state_prior, obs_prior, obs_incr,
                        state_h_dist, state_v_dist, state_t_dist,
                        hroi, vroi, troi,
                        h_local_func, v_local_func, t_local_func) -> None:
        return update_local_state_linear(state_prior, obs_prior, obs_incr,
                                         state_h_dist, state_v_dist, state_t_dist,
                                         hroi, vroi, troi,
                                         h_local_func, v_local_func, t_local_func)

    def update_local_obs(self, obs_data, used, obs_prior, obs_incr,
                         h_dist, v_dist, t_dist,
                         hroi, vroi, troi,
                         h_local_func, v_local_func, t_local_func) -> None:
        return update_local_obs_linear(obs_data, used, obs_prior, obs_incr,
                                       h_dist, v_dist, t_dist,
                                       hroi, vroi, troi,
                                       h_local_func, v_local_func, t_local_func)

@njit
def obs_increment_eakf(obs_prior, obs, obs_err) -> np.ndarray:
    nens = obs_prior.size

    ##obs error variance
    obs_var = obs_err**2

    ##obs_prior separate into mean+perturbation
    obs_prior_mean = np.mean(obs_prior)
    obs_prior_pert = obs_prior - obs_prior_mean

    ##compute prior error variance
    obs_prior_var = np.sum(obs_prior_pert**2) / (nens-1)

    """ensemble adjustment Kalman filter (Anderson 2003)"""
    var_ratio = obs_var / (obs_prior_var + obs_var)

    ##new mean is weighted average between obs_prior_mean and obs
    obs_post_mean = var_ratio * obs_prior_mean + (1 - var_ratio) * obs

    ##new pert is adjusted by sqrt(var_ratio), a deterministic square-root filter
    obs_post_pert = np.sqrt(var_ratio) * obs_prior_pert

    ##assemble the increments
    obs_incr = obs_post_mean + obs_post_pert - obs_prior

    return obs_incr

@njit
def update_local_state_linear(state_data, obs_prior, obs_incr,
                              h_dist, v_dist, t_dist,
                              hroi, vroi, troi,
                              h_local_func, v_local_func, t_local_func) -> None:

    nens, nfld, nloc = state_data.shape

    h_lfactor = h_local_func(h_dist, hroi)
    v_lfactor = v_local_func(v_dist, vroi)
    t_lfactor = t_local_func(t_dist, troi)

    nloc_sub = np.where(h_lfactor>0)[0]  ##subset of range(nloc) to update

    ##TODO: impact_on_state missing
    lfactor = np.zeros((nfld, nloc))
    for l in nloc_sub:
        for n in range(nfld):
            lfactor[n, l] = h_lfactor[l] * v_lfactor[n, l] * t_lfactor[n]

    state_data[:, :, nloc_sub] = update_ensemble(state_data[:, :, nloc_sub], obs_prior, obs_incr, lfactor[:, nloc_sub])

@njit
def update_local_obs_linear(obs_data, used, obs_prior, obs_incr,
                            h_dist, v_dist, t_dist,
                            hroi, vroi, troi,
                            h_local_func, v_local_func, t_local_func):

    ##distance between local obs_data and the obs being assimilated
    h_lfactor = h_local_func(h_dist, hroi)
    v_lfactor = v_local_func(v_dist, vroi)
    t_lfactor = t_local_func(t_dist, troi)

    lfactor = h_lfactor * v_lfactor * t_lfactor

    ##update the unused obs within roi
    ind = np.where(np.logical_and(~used, lfactor>0))[0]

    obs_data[:, ind] = update_ensemble(obs_data[:, ind], obs_prior, obs_incr, lfactor[ind])

@njit
def update_ensemble(ens_prior, obs_prior, obs_incr, local_factor) -> np.ndarray:
    nens = ens_prior.shape[0]
    ens_post = ens_prior.copy()

    ##obs-space statistics
    obs_prior_mean = np.mean(obs_prior)
    obs_prior_var = np.sum((obs_prior - obs_prior_mean)**2) / (nens-1)

    ##if there is no prior spread, don't update at all
    if obs_prior_var == 0:
        return ens_post

    cov = np.zeros(ens_prior.shape[1:])
    for m in range(nens):
        cov += ens_prior[m, ...] * (obs_prior[m] - obs_prior_mean) / (nens-1)

    reg_factor = cov / obs_prior_var

    ##the updated posterior ensemble
    for m in range(nens):
        ens_post[m, ...] = ens_prior[m, ...] + local_factor * reg_factor * obs_incr[m]

    return ens_post



import unittest
from NEDAS.models import registry, get_model_class, Model

class TestModelInterface(unittest.TestCase):
    def test_model_class_init(self):
        for model_name in registry.keys():
            ModelClass = get_model_class(model_name)
            model = ModelClass()
            self.assertIsInstance(model, Model)

if __name__ == '__main__':
    unittest.main()



import importlib
from NEDAS.config import Config
from .base import Assimilator

# make sure keys are lower-case
registry = {
    'ETKF': 'ETKFAssimilator',
    'EAKF': 'EAKFAssimilator',
    'TopazDEnKF': 'TopazDEnKFAssimilator',
    #'PDAF': 'PDAFAssimilator',
    #'RHF'
}

def get_assimilator(c: Config) -> Assimilator:
    """
    Get the correct Assimilator subclass instance based on the configuration.

    Args:
        c (Config): Configuration object.

    Returns:
        Assimilator: Corresponding Assimilator subclass instance.
    """
    if 'type' not in c.assimilator_def.keys():
        raise KeyError("'type' needs to be specified in assimilator_def")
    assimilator_type = c.assimilator_def['type']

    if assimilator_type not in registry:
        raise NotImplementedError(f"Assimilator type '{assimilator_type}' is not implemented")

    module = importlib.import_module('NEDAS.assim_tools.assimilators.'+assimilator_type)
    AssimilatorClass = getattr(module, registry[assimilator_type])

    return AssimilatorClass(c)



"""This module is used to perturb the forcing variables for the next forecast cycle.

This module is specifically designed for neXtSIM-DG in NEDAS.
The design follows the perturbation strategy in TOAPZ4 where
the perturbation is temporally correlationed as an AR1 process.

Parameters of the perturbation are read from the yaml file.
The perturbation is applied to the forcing variables in the forcing files.
"""
from datetime import datetime, timedelta, timezone
from dateutil.relativedelta import relativedelta # type: ignore
import os
import threading
import typing

import cftime # type: ignore
import netCDF4 # type: ignore
import numpy as np
import pyproj # type: ignore

from NEDAS.utils.conversion import t2s
from NEDAS.grid import Grid
from NEDAS.models.nextsim.dg.perturb import gen_perturb, apply_perturb, pres_adjusted_wind_perturb, apply_AR1_perturb
from NEDAS.models.nextsim.dg import slicing_nc

# both the topaz and ERA5 data are projected onto the same grid at the moment
_proj:pyproj.Proj = pyproj.Proj(proj='stere', a=6378273, b=6356889.448910593, lat_0=90., lon_0=-45., lat_ts=60.)
# the thread lock for reading and writing netcdf files
thread_lock = threading.Lock()

def get_fname_daterange(current_date: datetime, initial_date:str, interval:str, forcing_file_date_format:str) -> tuple[str, str]:
    """Inferring the date range of the forcing file for the current time,
    the forcing start date and the forcing interval in the initial forcing file given by yaml file.

    Parameters
    ----------
    current_date : datetime
        current date
    initial_date : str
        forcing start date
    interval : str
        forcing interval
    forcing_file_date_format : str
        forcing file date format expressed in strftime format

    Returns
    -------
    tuple[str, str]
        start date and end date of the forcing file for current time
    """
    # Parse the dates
    initial_date_dt:datetime = datetime.strptime(initial_date, forcing_file_date_format)
    initial_date_dt = initial_date_dt.replace(tzinfo=timezone.utc)
    keywords:dict[str, str] = {'y': 'years', 'm': 'months', 'd': 'days'}
    # Initialize start and end dates
    start_date = initial_date_dt
    end_date = initial_date_dt + relativedelta(**{keywords[interval[-1]]: int(interval[:-1])})

    assert current_date >= start_date, \
        f'Current time {current_date} is earlier than the initial forcing date {initial_date}'
    # Calculate the intervals until the current date is within the range
    while end_date <= current_date:
        start_date = end_date
        end_date = start_date + relativedelta(**{keywords[interval[-1]]: int(interval[:-1])})

    # Format the dates back to strings
    start_date_str = start_date.strftime(forcing_file_date_format)
    end_date_str = end_date.strftime(forcing_file_date_format)

    return start_date_str, end_date_str


def get_time_from_nc(fname:str, time_varname:str, time_units_name:str, time: datetime, next_time: datetime, debug:bool=False) -> tuple[np.ndarray, list[datetime]]:
    """Get the indices and corresponding time that includes time and next_time from the netcdf file

    This function is not seeking the exact time and next_time in the forcing file,
    but the time steps that include the time and next_time such that the perturbed forcing
    file can be a bit smaller. Therefore, we allow for a few more time steps in this file.

    Parameters
    ----------
    fname : str
        forcing file name
    time_varname : str
        time variable name in the forcing file
    time_units_name : str
        variable name that gives time units in the forcing file
    time : datetime
        current time
    next_time : datetime
        time at the end of the forecast cycle

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        indices and corresponding time for the next forecast cycle
    """
    with thread_lock:
        with netCDF4.Dataset(fname, 'r') as f:
            time_units = f[time_units_name].units
            # get the start time in the forcing file
            start_time: datetime = cftime.num2date(f[time_varname][0], units=time_units, only_use_cftime_datetimes=False)
            # get the time step in the forcing file
            time_step: timedelta = cftime.num2date(f[time_varname][1], units=time_units, only_use_cftime_datetimes=False) - start_time
            start_time = start_time.replace(tzinfo=timezone.utc)
            # get the indices of the current and next time steps
            it0: int = int(np.rint((time - start_time) / time_step))
            it1: int = int(np.rint((next_time - start_time) / time_step))
            # get total number of time steps in the current file
            nt: int = len(f[time_varname][:])
            # extend the forcing time step by one to ensure all time steps are included
            it0 = max(0, min(it0 - 1, nt - 1))
            it1 = max(0, min(it1 + 1, nt - 1))
            # get the all the time between current time and next time in the forcing file
            file_time: list[datetime] = [cftime.num2date(f[time_varname][it], time_units)
                                         for it in range(it0, it1 + 1)]
            if debug:
                print (f'file: {fname}; 'f'file time: {file_time[0]} to {file_time[-1]},'
                       f'forecast time: {time} to {next_time}')
    return np.arange(it0, it1 + 1), file_time

def get_time_index(fname:str, time_varname:str, time_units_name:str, time:datetime) -> int:
    """
    Get the index of time in a netcdf file
    """
    with thread_lock:
        with netCDF4.Dataset(fname, 'r') as f:
            time_units = f[time_units_name].units
            start_time: datetime = cftime.num2date(f[time_varname][0], units=time_units, only_use_cftime_datetimes=False)
            time_step: timedelta = cftime.num2date(f[time_varname][1], units=time_units, only_use_cftime_datetimes=False) - start_time
            start_time = start_time.replace(tzinfo=timezone.utc)
            ind: int = int(np.rint((time - start_time) / time_step))
    return ind

def get_prev_time_from_nc(fname:str, time_varname:str, time_units_name:str, itime:int) -> datetime:
    """Get the previous time in the netcdf file before the start of the forecast cycle

    Parameters
    ----------
    fname : str
        forcing file name
    time_varname : str
        time variable name in the forcing file
    time_units_name : str
        variable name that gives time units in the forcing file
    itime : int
        current time index in the forcing file
    """
    with thread_lock:
        with netCDF4.Dataset(fname, 'r') as f:
            it :int = max(0, itime - 1)
            time_units:str = f[time_units_name].units
            prev_time: datetime = cftime.num2date(f[time_varname][it], units=time_units, only_use_cftime_datetimes=False)
            prev_time = prev_time.replace(tzinfo=timezone.utc)
    return prev_time


def read_var(fname:str, varnames:list[str], itime: int) -> np.ma.MaskedArray:
    """reading a variable from a netcdf file

    Parameters
    ----------
    fname : str
        forcing file name
    varnames : list[str]
        list of variable names
    itime : int
        time index in the forcing file
    """
    data: list[np.ndarray] = []
    with thread_lock:
        with netCDF4.Dataset(fname, 'r') as f:
            # read the variable
            for vname in varnames:
                data.append(f[vname][itime])
    return np.ma.array(data)


def write_var(fname:str, varnames: list[str], data: np.ndarray, itime: int) -> None:
    """Write the perturbed variable back to the forcing file

    Parameters
    ----------
    fname : str
        forcing file name
    varnames : list[str]
        list of variable names
    itime : int
        time index in the forcing file
    """
    # We assume all variables in the forcing file exists
    assert os.path.exists(fname), f'{fname} does not exist; Please copy the forcing file to the correct path first.'
    with thread_lock:
        with netCDF4.Dataset(fname, 'r+') as f:
            for i, vname in enumerate(varnames):
                f[vname][itime] = data[i]
                f.sync()


def geostrophic_perturb(fname:str, grid:Grid, options:dict, itime:int, pert:np.ndarray, varname:str) -> None:
    """Perturb the atmosphere wind by the geostrophic balance.
    This applies to horizontal 2D wind fields.

    Parameters
    ----------
    fname : str
        forcing file name
    options : dict
        perturbation options for the geostrophic_wind_adjust section of atmosphere forcing from yaml file
    itime : int
        current time index in the forcing file
    pert : np.ndarray
        perturbation array
    varname : str
        name of the variable to be used to perturb wind fields

    Returns
    -------
    None
    """
    if not options['do_adjust']: return

    pres_name: str = options['pres_name']
    if pres_name != varname: return

    # doing wind perturbations by considering the geostrophic balance
    pert_u, pert_v = pres_adjusted_wind_perturb(grid,
                                                float(options['pres_pert_amp']),
                                                float(options['wind_pert_amp']),
                                                float(options['hcorr']), pert)

    uname:str = options['u_name']
    vname:str = options['v_name']
    u: np.ndarray = read_var(fname, [uname,], itime)
    v: np.ndarray = read_var(fname, [vname,], itime)
    u = apply_perturb(grid, u, pert_u, options['type'])
    v = apply_perturb(grid, v, pert_v, options['type'])
    if options['wind_amp_name'] != 'None':
        wind_amp_name:str = options['wind_amp_name']
        wind_amp = np.sqrt(u**2 + v**2)
        write_var(fname, [wind_amp_name,], wind_amp, itime)
    write_var(fname, [uname, ], u, itime)
    write_var(fname, [vname, ], v, itime)


def get_forcing_filename(forcing_file_options:dict, i_ens:int, time:datetime) -> str:
    """Get the forcing file name based on the current time and the forcing file format

    Parameters
    ----------
    forcing_file_options : dict
        forcing file options in the `file` section of the subsections of `perturb` section from the yaml file
    i_ens : int
        ensemble index
    time : datetime
        current time

    Returns
    -------
    str
        forcing file name
    """
    # derive the forcing file name
    # the format of the forcing file name
    file_format:str = forcing_file_options['format']
    # the date of the first forcing file
    forcing_file_initial_date:str = forcing_file_options['initial_date']
    # the length of the each forcing file
    forcing_file_interval:str = forcing_file_options['interval']
    # the length of the each forcing file
    forcing_file_date_format:str = forcing_file_options['datetime_format']
    # get the forcing file time
    forcing_start_date:str
    forcing_end_date:str
    forcing_start_date, forcing_end_date = \
        get_fname_daterange(time, forcing_file_initial_date, forcing_file_interval, forcing_file_date_format)
    fname: str
    try:
        fname = file_format.format(i=i_ens , start=forcing_start_date, end=forcing_end_date)
    except KeyError:
        try:
            fname = file_format.format(start=forcing_start_date, end=forcing_end_date)
        except KeyError:
            print ('Currently, we only supports keyword of 1. "start"+"end",'
                   '2. "start"+"end"+"i".'
                   'See the example yaml file for more information. '
                   'Modified the code if you have other requirements.')
    return fname


def perturb_forcing(forcing_options:dict, file_options:dict, i_ens: int, time: datetime, next_time:datetime, debug=False) -> None:
    """perturb the forcing variables

    Parameters
    ----------
    forcing_options : dict
        perturbation options from the yaml file
    file_options : dict
        forcing file options in the corresponding subsection of the `file` section from the yaml file
        e.g., info in the file/forcing/atmosphere is used in the perturb/forcing/atmosphere section
        Before calling this function, one must add the following keys to the file_options dictionary:
        - fname: the exact filename of the perturbed forcing file under absbolute path

    i_ens : int
        ensemble index
    time : datetime
        current time as the begining of the forecast cycle
    next_time : datetime
        end time of the next forecast cycle
    """
    # perturbation arrays
    pert: np.ndarray[typing.Any, np.dtype[np.float64]]
    # path to the directory of the perturbation files
    pert_path: str = forcing_options['path']
    # create the directory if it does not exist
    os.makedirs(os.path.join(pert_path, f'ensemble_{i_ens}'), exist_ok=True)
    # time index and time array
    time_index: np.ndarray[typing.Any, np.dtype[np.int64]]
    time_array: list[datetime]

    for forcing_name in forcing_options:
        if forcing_name not in file_options: continue
        # forcing options for each component, e.g., atmosphere or ocean
        forcing_options_comp:dict = forcing_options[forcing_name]
        file_options_comp:dict = file_options[forcing_name]
        # get the forcing file name
        fname:str = file_options_comp['fname']
        # copy forcing files to the ensemble member directory
        # we don't change the filename,
        # but only copy limited time slices of the original forcing file
        time_index, time_array = get_time_from_nc(file_options_comp['fname_src'],
                                                  file_options_comp['time_name'],
                                                  file_options_comp['time_units_name'],
                                                  time, next_time, debug
                                                  )
        # get prev_time
        prev_time:datetime = get_prev_time_from_nc(file_options_comp['fname_src'],
                                                   file_options_comp['time_name'],
                                                   file_options_comp['time_units_name'],
                                                   time_index[0]
                                                   )
        with thread_lock:
            slicing_nc.copy_time_sliced_nc_file(file_options_comp['fname_src'],
                                                fname, time_index,
                                                file_options_comp['time_name'],
                                                time_array[0])

        # get grid object for geometric information
        with thread_lock:
            with netCDF4.Dataset(fname, 'r') as f:
                grid = Grid(_proj, *_proj(f[file_options_comp['lon_name']],
                                        f[file_options_comp['lat_name']]
                                        )
                            )

        for itime, time_f in enumerate(time_array):
            # get options for perturbing the forcing variables
            options = forcing_options_comp['variables']
            for i, varname in enumerate(options['names']):
                if typing.TYPE_CHECKING:
                    assert type(varname) == str, 'variable name must be a string'
                # variable name in saved .npy filename
                varname_f:str = varname.replace("/", "_")
                # get perturbations
                pert_fname:str = os.path.join(pert_path, f'ensemble_{i_ens}',
                                              f'perturb_{varname_f}_{t2s(time_f)}.npy')
                if os.path.exists(pert_fname):
                    pert = np.load(pert_fname)
                else:
                    # convert the horizontal correlation length scale to grid points
                    hcorr:int = np.rint(float(options['hcorr'][i])/grid.dx)
                    # generate new perturbations
                    if prev_time != time_f:
                        prev_perturb_fname:str = os.path.join(pert_path,
                                                          f'ensemble_{i_ens}',
                                                          f'perturb_{varname_f}_{t2s(prev_time)}.npy')
                        pert_prev:np.ndarray = np.load(prev_perturb_fname)
                        # generate random perturbations for the current time step with AR1 correlation
                        pert_new:np.ndarray = gen_perturb(grid,
                                                          options['type'][i], float(options['amp'][i]),
                                                          hcorr)
                        pert = apply_AR1_perturb(pert_new, float(options['tcorr'][i]), pert_prev)
                    else:
                        pert = gen_perturb(grid, options['type'][i], float(options['amp'][i]), hcorr)

                    # save the perturbations for the next time step
                    perturb_file:str = os.path.join(pert_path,
                                                    f'ensemble_{i_ens}',
                                                    f'perturb_{varname_f}_{t2s(time_f)}.npy')
                    np.save(perturb_file, pert)

                # apply perturbations to the variable data
                # in the case of vector fields, we need to split the variable name
                varname_list: list[str] = varname.split(';')
                # read the variable data from forcing file
                data: np.ndarray = read_var(fname, varname_list, itime)
                # apply perturbations to the variable data
                data = apply_perturb(grid, data, pert, options['type'][i])
                # apply lower and upper bounds
                lb = float(options['lower_bounds'][i]) if options['lower_bounds'][i] != 'None' else -np.inf
                ub = float(options['upper_bounds'][i]) if options['upper_bounds'][i] != 'None' else np.inf
                data = np.minimum(np.maximum(data, lb), ub)
                # write the perturbed variable back to the forcing file
                write_var(fname, varname_list, data, itime)

                # generate wind perturbations and apply them to the atmosphere forcing files based on pressure perturbations
                if forcing_name == 'atmosphere': geostrophic_perturb(fname, grid,
                                                                     forcing_options_comp['geostrophic_wind_adjust'],
                                                                     itime, pert, varname)

            prev_time = time_f



import numpy as np
import typing
from scipy.optimize import fsolve, root_scalar # type: ignore
from scipy.ndimage import distance_transform_edt, gaussian_filter # type: ignore
from NEDAS.utils.spatial_operation import gradx, grady, warp
from NEDAS.utils.fft_lib import fft2, ifft2, get_wn, fftwn
from NEDAS.grid import Grid

def gen_perturb(grid: Grid,
                perturb_type: typing.Literal['gaussian'],
                amp: float,
                hcorr: int) -> np.ndarray:
    """
    get random perturbation fields

    Parameters
    ----------
    grid : Grid
        grid object for the 2D domain
    perturb_type : typing.Literal['gaussian']
        type of perturbation method
    amp : float
        amplitude of the perturbation
    hcorr : float
        horizontal decorrelation length scale in grid points
    powerlaw : float, optional
        power law exponent, by default -3.

    Returns
    -------
    np.ndarray
        random perturbation field
    """
    ##generate perturbation according to prescribed parameters
    if perturb_type == 'gaussian':
        perturb = random_field_gaussian(grid.nx, grid.ny, amp, hcorr)
        perturb = perturb - perturb.mean()
    else:
        raise TypeError('unknown perturbation type: '+perturb_type)

    return perturb


def apply_AR1_perturb(perturb: np.ndarray,
                      tcorr:float=1.0,
                      prev_perturb:typing.Union[None, np.ndarray]=None) -> np.ndarray:
    """apply AR1 perturbation to the field

    Parameters
    ----------
    perturb : np.ndarray
        perturbation field
    tcorr : float, optional
        time correlation length, by default 1.0
    prev_perturb : typing.Union[None, np.ndarray], optional
        previous perturbation field, by default None
    """
    ##create perturbations that are correlated in time
    autocorr = np.exp(-1.0)
    alpha = autocorr**(1.0/tcorr)
    if prev_perturb is not None:
        perturb = np.sqrt(1-alpha**2) * perturb + alpha * prev_perturb
    return perturb


def apply_perturb(grid: Grid,
                  field:np.ndarray,
                  perturb:np.ndarray,
                  perturb_type:typing.Literal['gaussian']) -> np.ndarray:
    """apply perturbation to the field

    Parameters
    ----------
    grid : Grid
        grid object for the 2D domain
    field : np.ndarray
        field to be perturbed
    perturb : np.ndarray
        perturbation field
    perturb_type : typing.Literal['gaussian']
        type of perturbation method

    Returns
    -------
    np.ndarray
        perturbed field
    """
    ##apply the perturbation
    # todo: adding logorithmic perturbation
    field += perturb
    return field


def random_field_gaussian(nx:int, ny:int, amplitude:float, hcorr:int) -> np.ndarray:
    """generate a 2D random field with Gaussian correlation length scale with given amplitude.

    Parameters
    ----------
    nx : int
        number of grid points in x direction
    ny : int
        number of grid points in y direction
    amplitude : float
        amplitude of the random field
    hcorr : int
        horizontal correlation length scale in grid points

    Returns
    -------
    np.ndarray
        2D random field with Gaussian correlation length scale
    """
    assert hcorr < nx, f'hcorr {hcorr} must be smaller than nx {nx}'
    assert hcorr > 0, f'hcorr {hcorr} must be larger than 0'
    # generate a 2D field with given spatial resolution
    fld: np.ndarray = np.zeros((ny, nx))
    # generate the wave frequency
    kappa:float = 2*np.pi/(nx)
    lamda:float = 2*np.pi/(ny)
    # squared wave frequency
    kappa2:float = kappa*kappa
    lambda2:float = lamda*lamda
    # get the wave numbers
    n1:np.ndarray = fftwn(nx)
    n2:np.ndarray = fftwn(ny)
    l: np.ndarray
    p: np.ndarray
    l, p = np.meshgrid(n1, n2)

    def func2d(sigma:float) -> float:
      """solve the equation for sigma given horizontal correlations.

      This function is solved to obtain the sigma for unit variance of the random field.
      """
      sum1:float = np.sum(np.exp(-2.0*(kappa2*l*l+lambda2*p*p)/sigma/sigma)*np.cos(kappa*l*hcorr))
      sum2:float = np.sum(np.exp(-2.0*(kappa2*l*l+lambda2*p*p)/sigma/sigma))
      return sum1/sum2 - np.exp(-1)

    # solve for sigma given hcorr so that func2d=0
    sigma:float = root_scalar(func2d, bracket=(1e-6, 2*np.pi), method='bisect', xtol=1e-10).root

    # compute the scaling factor c for getting the unit variance
    sum2:float = np.sum(np.exp(-2.*(kappa2*l*l+lambda2*p*p)/sigma/sigma))
    c:float = 1./np.sqrt(sum2)

    # getting the random phase shift in Fourier space
    val:np.ndarray = np.random.uniform(0, 1, fld.shape)
    val = np.exp(2*np.pi*val*1j)

    # getting the Fourier space representation of the field
    gauss:np.ndarray = np.exp(-(kappa2*l*l+lambda2*p*p)/sigma/sigma)
    ph:np.ndarray = amplitude* c* gauss * val
    # the nx*ny is the normalisation factor because fftw normalise the inverse fft by default
    return np.real(ifft2(ph * nx * ny))


def pres_adjusted_wind_perturb(grid: Grid, ampl_pres:float, ampl_wind:float,
                               scorr:float, pres:np.ndarray,
                               with_wind_speed_limit:bool=True, wlat:float=15.) -> tuple[np.ndarray, np.ndarray]:
    """generate a random perturbation for wind u,v by pressure perturbation.

    Parameters
    ----------
    grid : Grid
        grid object for the 2D domain
    ampl_pres : float
        amplitude of pressure perturbations (Pa)
    ampl_wind : float
        amplitude of wind perturbations (m/s)
    scorr : float
        horizontal decorrelation length scale by number of grid points
    pres : np.ndarray
        pressure field perturbations
    with_wind_speed_limit : bool, optional
        if true: limit pressure perturbation by wind speed to account for horizontal scale of perturbation, by default True.
        This option will force `pres_wind_relate` to be True.
    wlat : float, optional
        latitude bound for pure geostroph wind, by default 15.

    Returns
    -------
    tuple
        pressure perturbation, u wind perturbation, v wind perturbation
    """

    # get latitude from grid
    plat: np.ndarray
    _, plat = grid.proj(grid.x, grid.y, inverse=True)
    # convert radians to degrees
    r2d:float = 180/np.pi
    # air density
    rhoa:float = 1.2
    # reference latitude where we use the coriolis parameter
    rlat:float = 40.
    # coriolis parameter where we use 40 degrees as the latitude
    # the sign function is used to distinguish the coriolis parameter for different hemisphere
    fcor:np.ndarray = np.sign(plat)*(2*np.sin(rlat/r2d)*2*np.pi/86400)

    # ratio between the wind speed amplitude and a typical wind speed from pressure gradient
    wprsfac:float =1.
    if with_wind_speed_limit:
        # typical pressure gradient
        wprsfac=np.sqrt(ampl_pres)/scorr/fcor
        wprsfac=np.sqrt(ampl_wind)/wprsfac

    # calc u,v from pres according to pres-wind relation
    # pres gradients
    dpresx:np.ndarray = gradx(pres, grid.dx)*wprsfac
    dpresy:np.ndarray = grady(pres, grid.dx)*wprsfac

    # geostrophic wind near poles
    vcor: np.ndarray =  dpresx / fcor / rhoa
    ucor: np.ndarray = -dpresy / fcor / rhoa

    # wind near equator is proportional to pressure gradient
    # coriolis parameter is just used to limit the equator wind perturbations
    ueq: np.ndarray = -dpresx / np.abs(fcor) / rhoa
    veq: np.ndarray = -dpresy / np.abs(fcor) / rhoa

    # weight for geostrophic wind
    # the weight is 1 when the latitude is outside of the band of wlat
    # the weight is 0 when the latitude is within the band of wlat
    wcor:np.ndarray = np.sin(np.minimum(wlat, np.abs(plat)) / wlat * np.pi * 0.5)
    u = wcor*ucor + (1-wcor)*ueq
    v = wcor*vcor + (1-wcor)*veq

    return u, v



import numpy as np
from scipy.interpolate import LinearNDInterpolator
from NEDAS.utils.njit import njit

@njit
def lonlat2sxy(lon, lat):
    """convert lon,lat to stereographic plane x,y"""
    invrad = np.pi / 180.
    r = np.cos((lat+90)/2*invrad)
    x = r * np.cos(lon*invrad)
    y = r * np.sin(lon*invrad)
    return x, y

@njit
def sxy2lonlat(sx, sy):
    """convert stereographic plane x,y to lon,lat"""
    rad = 180. / np.pi
    r = np.hypot(sx, sy)
    lat = 2 * rad * np.arccos(r) - 90
    lon = rad * np.arctan2(sy, sx)
    return lon, lat

@njit
def spherdist(lon1, lat1, lon2, lat2):
    """
    Calculate distance between (lon1,lat1) and (lon2,lat2)

    This function treats the Earth as a perfect sphere, it is
    not as accurate as pyproj.Geod.inv which takes the ellipse
    into account. But for the purpose of searching for positions
    in the domain, this function is accurate enough and faster.

    Inputs: lon1, lat1, lon2, lat2: float, degrees
    Output: dist: float, meters
    """
    invrad = np.pi / 180.
    Re = 6.371e6

    dlon = lon2 - lon1

    cosarc =  (np.sin(lat1*invrad) * np.sin(lat2*invrad) +
               np.cos(lat1*invrad) * np.cos(lat2*invrad) * np.cos(dlon*invrad))

    dist = Re * np.arccos(np.minimum(1., np.maximum(-1., cosarc)))

    return dist

@njit
def tri_area(a, b, c):
    """compute triangle area given edge lengths"""
    s = 0.5 * (a + b + c)
    d = s * (s-a) * (s-b) * (s-c)
    if d < 0:
        return 0  ##the triangle is not well defined
    else:
        return np.sqrt(d)

@njit
def pivotp(glon, glat, neighbors, lon, lat, ipiv=0, jpiv=0):
    """in model glon,glat, find the pivot point for lon,lat"""

    min_d = spherdist(glon[jpiv, ipiv], glat[jpiv, ipiv], lon, lat)
    while min_d > 0:
        i_c, j_c = ipiv, jpiv

        ##search in the 4 neighbors
        for n in range(4):
            j_n, i_n = neighbors[:, n, jpiv, ipiv]
            d = spherdist(glon[j_n, i_n], glat[j_n, i_n], lon, lat)
            if d < min_d:
                ipiv, jpiv = i_n, j_n
                min_d = d

        ##stop if no progress are made
        if ipiv==i_c and jpiv==j_c:
            break
    return ipiv, jpiv

@njit
def find_grid_index(glon, glat, gx, gy, neighbors, lon, lat, ipiv, jpiv):
    """ convert a point lon,lat to grid space x,y """

    pt = jpiv, ipiv
    pt_e = neighbors[0, 0, *pt], neighbors[1, 0, *pt]
    pt_n = neighbors[0, 1, *pt], neighbors[1, 1, *pt]
    pt_w = neighbors[0, 2, *pt], neighbors[1, 2, *pt]
    pt_s = neighbors[0, 3, *pt], neighbors[1, 3, *pt]
    pt_ne = neighbors[0, 1, *pt_e], neighbors[1, 1, *pt_e]
    pt_nw = neighbors[0, 1, *pt_w], neighbors[1, 1, *pt_w]
    pt_se = neighbors[0, 3, *pt_e], neighbors[1, 3, *pt_e]
    pt_sw = neighbors[0, 3, *pt_w], neighbors[1, 3, *pt_w]

    ##find which triangle (lon,lat) falls in
    ##  pt_nw --- pt_n ----- pt_ne
    ##     | .  3  .   2   .  \
    ##     | 4  .   .  .   1   \
    ##   pt_w...... pt......... pt_e
    ##     | 5   .   .   .   8   \
    ##     |  .   6   .  7    .   \
    ##  pt_sw ----- pt_s -------- pt_se

    pt_lst   = [[pt_e, pt_ne], [pt_ne, pt_n], [pt_n, pt_nw], [pt_nw, pt_w], [pt_w, pt_sw], [pt_sw, pt_s], [pt_s, pt_se], [pt_se, pt_e]]
    xoff_lst = [[   1,     1], [    1,    0], [   0,    -1], [   -1,   -1], [  -1,    -1], [   -1,    0], [   0,     1], [    1,    1]]
    yoff_lst = [[   0,     1], [    1,    1], [   1,     1], [    1,    0], [   0,    -1], [   -1,   -1], [  -1,    -1], [   -1,    0]]

    for p in range(len(pt_lst)):
        pt1 = pt
        pt2 = pt_lst[p][0]
        pt3 = pt_lst[p][1]
        xoff2 = xoff_lst[p][0]
        xoff3 = xoff_lst[p][1]
        yoff2 = yoff_lst[p][0]
        yoff3 = yoff_lst[p][1]

        d1 = spherdist(lon, lat, glon[pt], glat[pt])
        d2 = spherdist(lon, lat, glon[pt2], glat[pt2])
        d3 = spherdist(lon, lat, glon[pt3], glat[pt3])
        s1 = spherdist(glon[pt2], glat[pt2], glon[pt3], glat[pt3])
        s2 = spherdist(glon[pt3], glat[pt3], glon[pt], glat[pt])
        s3 = spherdist(glon[pt], glat[pt], glon[pt2], glat[pt2])
        A1 = tri_area(s1, d2, d3)
        A2 = tri_area(s2, d1, d3)
        A3 = tri_area(s3, d1, d2)
        Atot = tri_area(s1, s2, s3)
        if Atot == 0:  ##the triangle is not well defined
            continue

        ##find barycentric coordinates
        b1 = A1 / Atot
        b2 = A2 / Atot
        b3 = A3 / Atot
        sum_b = b1 + b2 + b3
        if np.abs(sum_b - 1) > 1e-3:   ##point falls outside this triangle
            continue

        ##the output is weighted average of x,y of triangle vertices
        x1, y1 = pt[1], pt[0]
        x2, y2 = x1+xoff2, y1+yoff2
        x3, y3 = x1+xoff3, y1+yoff3
        x = x1*b1 + x2*b2 + x3*b3
        y = y1*b1 + y2*b2 + y3*b3
        # print(ipiv, jpiv, ':', pt2, pt3, ':', b1, b2, b3, ':', x, y)

        return x, y

    return np.nan, np.nan

def lonlat2xy(glon, glat, gx, gy, neighbors, lon, lat):
    """ convert from lon,lat to grid space index x,y """
    x, y = np.full(lon.size, np.nan), np.full(lon.size, np.nan)
    ipiv, jpiv = 0, 0
    for i in range(lon.size):
        ipiv, jpiv = pivotp(glon, glat, neighbors, lon[i], lat[i], ipiv, jpiv)
        x[i], y[i] = find_grid_index(glon, glat, gx, gy, neighbors, lon[i], lat[i], ipiv, jpiv)
    return x, y

def xy2lonlat(glon, glat, gx, gy, neighbors, x, y):
    """ convert from grid space index x,y to lon,lat """
    ny, nx = glon.shape
    glon_, glat_ = np.zeros((ny+2, nx+2)), np.zeros((ny+2, nx+2))
    glat_[1:-1, 1:-1] = glat
    glat_[1:-1, 0] = glat[:, -1]  ##cyclic west boundary
    glat_[1:-1, -1] = glat[:, 0]  ##cyclic east boundary
    glat_[-1, :] = glat_[-2, ::-1]  ##north boundary
    glat_[0, :] = -80.2851  ##south boundary, add another latitude band
                            ##better: extrapolated from lat[1:5,:]
    glon_[1:-1, 1:-1] = glon
    glon_[1:-1, 0] = glon[:, -1]  ##cyclic west boundary
    glon_[1:-1, -1] = glon[:, 0]  ##cyclic east boundary
    glon_[-1, :] = glon_[-2, ::-1]  ##north boundary
    glon_[0, :] = glon_[1, :]   ##south boundary, add another latitude band

    gx_, gy_ = np.meshgrid(np.arange(-1, nx+1), np.arange(-1, ny+1))
    sx, sy = lonlat2sxy(glon_, glat_)
    xy2sx = LinearNDInterpolator(list(zip(gx_.flatten(), gy_.flatten())), sx.flatten())
    xy2sy = LinearNDInterpolator(list(zip(gx_.flatten(), gy_.flatten())), sy.flatten())

    lon, lat = sxy2lonlat(xy2sx(x, y), xy2sy(x, y))
    return lon, lat

    ##get 1d coordinates in x,y directions
    # ny, nx = gx.shape

    ##check if x,y is inside the grid
    ##allowed range is -1 to nx (extended 1 grid point to both sides)
    # if x>=-1 and x<=nx and y>=0 and y<=ny:

    #     i, j = int(np.floor(x)), int(np.floor(y))

    #     ##get the four vertices
    #     if i<0 and j<0:
    #         j2,i2 = j+1,i+1
    #         j1,i1 = neighbors[:,3,j2,i2]  ##point to the south
    #         j3,i3 = neighbors[:,2,j2,i2]  ##point to the west
    #         j0,i0 = neighbors[:,2,j1,i1]  ##point to the south-west

    #     elif i<0 and j>=0:
    #         j1,i1 = j,i+1
    #         j0,i0 = neighbors[:,2,j1,i1]  ##point to the west
    #         j2,i2 = neighbors[:,1,j1,i1]  ##point to the north
    #         j3,i3 = neighbors[:,1,j0,i0]  ##point to the north-west

    #     elif i>=0 and j<0:
    #         j3,i3 = j+1,i
    #         j0,i0 = neighbors[:,3,j3,i3]  ##point to the south
    #         j2,i2 = neighbors[:,0,j3,i3]  ##point to the east
    #         j1,i1 = neighbors[:,0,j0,i0]  ##point to the south-east

    #     else:
    #         j0,i0 = j,i
    #         j1,i1 = neighbors[:,0,j0,i0]  ##point to the east
    #         j3,i3 = neighbors[:,1,j0,i0]  ##point to the north
    #         j2,i2 = neighbors[:,1,j1,i1]  ##point to the north-east

    #     ##internal coordinates for interpolation weights
    #     u, v = x - i, y - j
    #     w = np.array([(1-u)*(1-v), u*(1-v), (1-u)*v, u*v])

    #     ##lon,lat at vertices, convert to stere proj, interp, convert back
    #     lon_pt = np.array([glon[j0,i0], glon[j1,i1], glon[j2,i2], glon[j3,i3]])
    #     lat_pt = np.array([glat[j0,i0], glat[j1,i1], glat[j2,i2], glat[j3,i3]])
    #     sx_pt, sy_pt = lonlat2sxy(lon_pt, lat_pt)
    #     sx, sy = np.sum(w*sx_pt), np.sum(w*sy_pt)
    #     lon, lat = sxy2lonlat(sx, sy)

    #     return lon, lat

    # else:
    #     return np.nan, np.nan

